;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;												
;; eForth for Rabbit 2000
;;
;; eForth is free for non-commercial use
;;
;;												
;;	KLS_WORD 	= WORD I ADDED							
;;	KLS		= MODIFICATION							
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Derived by combining my 68306 eForth with 
;; assembly words defined by Ken Chen for his Z80 eForth.

; 05/14/97		all DB strings must be divisible by 4! (cell aligned)
; 05/14/97		string compare on cell basis!
; 09/19/97		Add DOES> and CONSTANT
; 10/01/97		fix DUMP format ( U.R2 and dm+ )
; 10/06/97		SEES. Decompiler that stops at CALLN ( NOP JSR )
; 10/07/97		VOCABULARY. Fixed vocabulary useage.
; 10/21/97		Add ROMin to define variables and words in RAM at boot time
; 03/07/2000	Merge code and name space
; 03/20/2000	Re-wrote TOKEN & $,n for merged name/code 
; 03/20/2000	Fixed doDOES words (CONSTANT & VOCABULARY) for merged name/code
; 05/20/2000	Made Decompiled words for SREC address independent
; 10/14/2000	Fixed vocabulary useage again. Rev 2 compatible.

; 10/15/2001	Initial Rabbit 2000 coding
;			Added comments to Z80 assy source
; 10/20/2001	exit back to dyn-c
; 			baud rate set by parameter in HL
; 10/21/2001	P!, P@ i/o port access
; 10/23/2001	added INIT word search to boot
; 			added forget (restores base forth)
; 10/25/2001	flash, forget return with HL = HERE
;			bye returns with HL = 0.
; 10/30/2001	version format n.na n=#,a={a..z}
; 11/10/2001	2.0b added BLOAD. redefined LOAD.
; 04/15/2003	support for serial A,B,or C. define macro SERIAL.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; eForth 2.0 uses LOAD, READ to work around the flow control		
;; problem with compilation.  First load a file using text transfer.	
;; Then read the file (compile it).					
;;										
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;===============================================================
;
;	eForth 1.0 by Bill Muench and C. H. Ting, 1990
;	Much of the code is derived from the following sources:
;		8086 figForth by Thomas Newman, 1981 and Joe smith, 1983
;		aFORTH by John Rible
;		bFORTH by Bill Muench
;		Z80 eForth by Ken Chen
;
;	The goal of this implementation is to provide a simple eForth Model
;	which can be ported easily to many 8, 16, 24 and 32 bit CPU's.
;	The following attributes make it suitable for CPU's of the '90:
;
;		small machine dependent kernel and portable high level code
;		source code in the MASM format
;		direct threaded code
;		separated code and name dictionaries
;		simple vectored terminal and file interface to host computer
;		aligned with the proposed ANS Forth Standard
;		easy upgrade path to optimize for specific CPU
;
;	You are invited to implement this Model on your favorite CPU and
;	contribute it to the eForth Library for public use. You may use
;	a portable implementation to advertise more sophisticated and
;	optimized version for commercial purposes. However, you are
;	expected to implement the Model faithfully. The eForth Working
;	Group reserves the right to reject implementation which deviates
;	significantly from this Model.
;
;	As the ANS Forth Standard is still evolving, this Model will
;	change accordingly. Implementations must state clearly the
;	version number of the Model being tracked.
;
;	Representing the eForth Working Group in the Silicon Valley FIG Chapter.
;	Send contributions to:
;
;		Dr. C. H. Ting
;		156 14th Avenue
;		San Mateo, CA 94402
;		(415) 571-7639
;
;===============================================================


;; CONDITIONAL ASSEMBLY:
DIAGS		EQU	0	; NO DIAGNOSTICS
;DIAGS		EQU	1	; YES DIAGNOSTICS

;; Version control

VER		EQU	2			;major release version
EXT		EQU	0			;minor extension
ALFA		EQU	"e"			;alpha extension


CELLSIZE	EQU	2


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;												
;; Rabbit 2000 Macro definitions							
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

SERIAL	EQU	SPORTC
	
;;
;; define IOI opcode
;;
IOI     MACRO
        DB      0D3H
        ENDM
;;
;; port output
;;
POUT    MACRO   port,data
        LD      A,data
        IOI
        LD      (port),A
        ENDM

;;
;; port input
;;
PIN     MACRO   port
        IOI
        LD      A,(port)
        ENDM


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;												
;; eForth Macro definitions								
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;
;; cell alignment
;;
M_ALIGNC        MACRO
        align CELLSIZE
        ENDM

;;
;; Assemble inline direct threaded code ending.
;;
;; the forth instruction pointer (IP) in BC is pointing to the to-be-executed one
M_NEXT  MACRO
        LD      A,(BC)		;put low byte of next instr in A
        INC     BC		;inc instr ptr
        LD      L,A		;move next instr low byte to L
        LD      A,(BC)		;put high byte of next instr in A
        INC     BC		;inc instr ptr
        LD      H,A		;move next instr high byte to H
        JP      (HL)		;jump to next instruction
        ENDM

;;
;; Compile a colon definition header
;;
M_COLON MACRO   LEX,NAMED,LBLNAME
        align   CELLSIZE
        DW      LBLNAME,LINK
LINK    EVAL $
        DB      LEX,NAMED
        align   CELLSIZE

LBLNAME label $
        DW      CALLL
        DW      DOLST

        ENDM

;;
;; Compile a code definition header
;;
M_CODE  MACRO   LEX,NAMED,LBLNAME
        align   CELLSIZE
        DW      LBLNAME,LINK
LINK    EVAL $
        DB      LEX,NAMED
        align   CELLSIZE

LBLNAME label $
        ENDM

;;
;; Compile a user variable header
;;
M_USER  MACRO   LEX,NAMED,LBLNAME
        align   CELLSIZE
        DW      LBLNAME,LINK
LINK    EVAL $
        DB      LEX,NAMED
        align   CELLSIZE

LBLNAME label $
        DW      CALLL
        DW      DOLST
        DW      DOUSE,USERV
USERV   EVAL USERV+CELLL
        ENDM


;kls; DDOT:                EQU     DOTS    ;debug .S (for easy deletion) KLS


;* R2000 System Registers
; timer A
TACSR	equ	0A0H
TACR	equ	0A4H
TAT6R	equ	0ADH		; serial c
TAT5R	equ	0ABH		; serial b
TAT4R	equ	0A9H		; serial a
; port E
PEDR	equ	070H
PEFR	equ	075H
PEDDR	equ	077H
PECR	equ	074H
; port C
PCFR	equ	055H
PCDR	equ	050H
;; serial port offsets
STATREG	equ	003H
CTLREG	equ	004H
; serial C
SPORTC	equ	0E0H
SCSR	equ	0E3H
SCCR	equ	0E4H
SCDR	equ	0E0H
; serial B
SPORTB	equ	0D0H
SBSR	equ	0D3H
SBCR	equ	0D4H
SBDR	equ	0D0H
; serial A
SPORTA	equ	0C0H
SASR	equ	0C3H
SACR	equ	0C4H
SADR	equ	0C0H			
; global clock
GCDR	equ	 00FH
; watch dog
WDTCR	equ	008H
WDTTR	equ	009H
; rtc
RTCCR	equ	001H
RTC0R	equ	002H
RTC1R	equ	003H
RTC2R	equ	004H
RTC3R	equ	005H
RTC4R	equ	006H
RTC5R	equ	007H
; global control/status
GCRS	equ	000H



;*************************************************************************** 


;  ROM $0-romsize. 128K flash on rcm2300.
;ROML	EQU	0x0000
;ROMH	EQU	0x5FFF+1
; runs from ram. gets relocated by dynamic c.

;  RAM 128K ram on rcm2300.
RAML	EQU	00000H		; buffer start
RAMH	EQU	0AFFFH+1	; stay out of stack space (& debug area)!
;;RAMH	EQU	0xDFFF+1	; stack space not used after eforth launch
				; this may change if i try to return to
				; dyn-c code to write flash

STACKSIZE	EQU	00800H	; 2K stack builds DOWN from RAMH

;ROMSIZE	EQU	ROMH-ROML


	ORG    RAML


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

main:
	push	bc
	push	de
	push	hl
	push	af
	push	iy
	push	ix		; saves state on dyn-c stack
	ld	(SAVE_SP),sp	; remember where dyn-c stack was

	ld	A,L
	ld	(bauddiv),A	; bauddiv is parameter

	ld	A,H
	ld	(port),A	; port is parameter

;; 	jp	INITE
	
;(*** KLS ***) 

;
; initialize Rabbit 2000
;; HARDWARE SPECIFIC
;

	POUT WDTTR,053H
	POUT WDTTR,054H			; NO WATCHDOG!

; done in bios...	
;//	 POUT(GCDR,000H)		; NO DOUBLER
;//	 POUT(GCR,008H)			; sysclk & perclk = osc. no ints.

    POUT PCFR,044H			; Select TxA & TxC (i/o port c alternate function)

	POUT (SERIAL+CTLREG),000H		;  serial port control:	portc, 8bit aync, no int

; timer a stuff    
	LD A,(bauddiv)		

	IOI
	LD (TAT6R),A			; set timer A6 to baud divisor
	nop

	IOI
	LD (TAT5R),A			; set timer A5 to baud divisor
	nop

	IOI
	LD (TAT4R),A			; set timer A4 to baud divisor
	nop

    POUT TACR,000H			; ALL pclk/2. no ints.
    POUT TACSR,001H			; enable timer A

INITE:	
; port e stuff
	POUT PECR,000H		; pclk/2 transfer
	POUT PEFR,000H		; normal i/o
	POUT PEDDR,080H		; e7 is output. led turns on.

; delay for initialization
;;// let baud timer time out before using port!
	LD	A,010H
L006:	LD	DE,00000H
	LD	HL,00000H
L005:	INC	HL
	DB	0ECH		;	OR	HL,DE	(new rabbit 2000 instruction) (kls)
	JR	NZ,L005
	DEC	A
	JR	NZ,L006

; boot message
	LD	HL,MSG1
L003:	LD	C,(HL)
	INC	HL
L004:
	PIN	((SERIAL+STATREG))        ; get status

	BIT     2,A         ; tx busy?
	JR      NZ,L004   ; loop until its not (just serial shift delay)
	LD      A,C

	IOI 
	LD  	(SERIAL),A        ; send the character
	NOP			;BURNED BY THE INSTRUCTION BUG

	LD	A,(HL)
	OR	A		; set Z status
	JR	NZ,L003

	
;(***KLS***) (*END INIT CODE*)

	JP	START_FORTH



;***** MESSAGE AREA *****
	DW	0

MSG1  DW    0A0DH
      DW    0A0DH
	DB	"Rabbit OK"
	DW	0A0DH
	DB	"Booting eForth..."
	DW	0A0DH
MSG1E DW    0000H


MSG2  DW	0A0DH
      DW	0A0DH
	DB	"Breakpoint Reached"
	DW	0A0DH
MSG2E DW	0000H


; ***** END OF MESSAGE AREA. ***** BE SURE WE'RE STILL ALIGNED.

	align CELLSIZE

; STICK MEMORY REGISTERS FOR RETURN PTR & USER PTR HERE...

RPr	DW	0000H
UPr	DW	0000H

; STICK DYN-C SP here
; with parameters passed in HL

SAVE_SP	DW	0000H
bauddiv	DW	0000H
port	DW	0000H	
	

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;												
;; BEGIN EFORTH CODE...									
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Rabbit 2000 Registers used in eForth model
;
;	Registers assigment:
;	SP  --  SP 			Data Stack Pointer, SP
;	IP  --  BC			Interpreter Pointer, IP
;	RPr  --  Memory (sim reg)	Return Stack Pointer, RP
;	UPr  --  Memory (sim reg)	User Pointer, UP
;
; 		Other Register uses:
;		HL,BC,DE		UM* (code)
;		IX			rominit index
;		A			scratch
;		HL			code scratch
;		DE			code scratch @,!
;

;; Constants

TRUEE		EQU	-1			;true flag

COMPO		EQU	040H			;lexicon compile only bit
IMEDD		EQU	080H			;lexicon immediate bit
;; IMEDD and COMPO are added to length byte as part of DB statement
MASKK		EQU	07F1FH			;lexicon bit mask

CELLL		EQU	2			;size of a cell - 16-bit Forth
BASEE		EQU	16			;default radix
VOCSS		EQU	8			;depth of vocabulary stack

BKSPP		EQU	8			;back space
LF		EQU	10			;line feed
CRR		EQU	13			;carriage return
ERR		EQU	27			;error escape
TIC		EQU	39			;tick

CALLL		EQU	0CD00H			;Z80 nop call opcodes 
						; byte reversed when DWL CALLL


;; Memory allocation
;; Memory allocation	0//code/name>--//up>--<sp//tib>--rp//em

;; RAM allocation:
;;
;;	top of RAM (ramend)	+-------------------------------+
;;			EM	SPP	|   data stack grows down	  |
;;					|            \/			  | SS
;;					|            \/		        |
;;					+-------------------------------+
;;					|            /\                 |
;;					|            /\			  | US
;;				UPP	|    user area grows up		  |
;;					+-------------------------------+
;;				RPP	|    return stack grows down	  |
;;					|            \/			  |
;;					|            \/			  |
;;					+-------------------------------|
;;					|            /\                 |
;;					|            /\			  |
;;				TIB	|   text input buffer grows up  |
;;					+-------------------------------+
;;					|            /\                 |
;;					|            /\                 |
;;					|            /\			  |
;;				CP NP	|   code & name                 |
;;					|   dictionary grows up 	  |
;;					|				        |
;;				CODEE	|   eforth dictionary		  |
;;	bottom of RAM (rambeg)	+-------------------------------+
;;
;;
;;
;; DICTIONARY entries build low memory to high memory
;;
;;
;; $n100	Code Addr.L
;; $n104	Link Addr.L	(points to $0000)
;; $n108	Lex & Length BYTE
;; $n109	"B"
;; $n10A	"Y"
;; $n10B	"E"
;; $n10C	code for word...
;; 		....
;; $n120	Code Addr.L				CFA
;; $n124	Link Addr.L	(points to $n108)	LFA
;; $n128	Lex & Length				NFA
;; $n129	"?"
;; $n12A	"R"
;; $n12B	"X"
;; $n12C	code for word ...			PFA
;;		...
;; CFA Code Field Address
;; LFA Link Field Address
;; NFA Name Field Address
;; PFA Parameter Field Address
;;
;;  where Lex is immed if this is an IMMEDIATE word and compo if
;;  this is a COMPILE-ONLY word.
;; 
;; CODE and NAME entries are created with the following macros...
;;
;; CELL alignment occurs with _NAME arithmetic & ORG
;; length of name string is included in this arithmetic
;;



EM		EQU	RAMH			;top of RAM
;BM		EQU	ROML			;bottom of ROM
SS		EQU	STACKSIZE		;data stack size
US		EQU	64*CELLL		;user area size in cells
RTS		EQU	128*CELLL		;return stack/TIB size

;SPP		EQU	TIBB-8*CELLL		;start of data stack (SP0)
SPP		EQU	RAMH			;start of data stack (SP0)
UPP		EQU	EM-SS-US		;start of user area (UP0)
RPP		EQU	UPP-8*CELLL		;start of return stack (RP0)
TIBB		EQU	RPP-RTS			;terminal input buffer (TIB)



;; Initialize assembly _VARIABLES_

LINK	EVAL 0					;force a null link
USERV	EVAL 4*CELLL				;first user variable offset (SPP)




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; Merging CODE & NAME space affects words that use NTOP/NP and CTOP/CP	
;; NTOP/NP:	TOKEN	$,n								
;; CTOP/CP:	HERE	ALLOT	,	$,"						
;; so change TOKEN and $,n								
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;
; CODE word for breakpoint.
; Insert in CODE word lists for debugging.
;

	IF	0		

	M_CODE	3,"bpt",bpt
;*	GET START ADDRESS OF THE MESSAGE
	LD	HL,MSG2
bptL1:	LD	C,(HL)
	INC	HL
bptL2:
	PIN	((SERIAL+STATREG))	; get status
	BIT     2,A			; tx busy?
	JR      NZ,bptL2   ; loop until its not (just serial shift delay)
	LD      A,C
	IOI 
	LD  	(SERIAL),A	; send the character
	NOP				;BURNED BY THE INSTRUCTION BUG
	LD	A,(HL)
	OR	A			; set Z status
	JR	NZ,bptL1

bptL3:	jp	bptL3			; STOP HERE! (NO M_NEXT)

	ENDIF


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; Main entry points and COLD start data						
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;MAIN

	M_ALIGNC	; FORTH MUST BE CELL ALIGNED...

START_FORTH:


; init SP, RPr, UPr
		LD  HL,SPP
		LD  SP,HL    ; INIT SP
                
		LD   HL,RPP   ;   init to RPP
		LD   (RPr),HL  ;   simulated R-stack reg at RP
                
		LD   HL,UPP    ; use the initial USER VARIABLE area
		LD   (UPr),HL   ; simulated R-stack reg at UP
                
;		LD  BC, START_FORTH   ; re-start whenever R-stack underflow-used
		LD  BC,COLD1
		M_NEXT

		JP  START_FORTH		; just in case




; COLD start moves the following to USER variables.
; MUST BE IN SAME ORDER AS USER VARIABLES.
; USER variables are in RAM and get initialized
; with these values.

;;	ALGN WORDSIZE				;align to CELL boundary
		M_ALIGNC

UZERO:	DW	0,0,0,0		;reserved space in user area
		DW	SPP			;SP0 			4
		DW	RPP			;RP0 			5
		DW	QRX			;'?KEY 		6
		DW	TXSTO			;'EMIT 		7
		DW	ACCEP			;'EXPECT 		8
		DW	KTAP			;'TAP	 		9
		DW	TXSTO			;'ECHO 		10
		DW	DOTOK			;'PROMPT 		11
		DW	BASEE			;BASE 		12
		DW	0			;tmp 			13
		DW	0			;SPAN 		14
		DW	0			;>IN 			15
		DW	0			;#TIB 		16
		DW	TIBB			;TIB 			17
		DW	0			;CSP 			18
		DW	INTER			;'EVAL		19
		DW	NUMBQ			;'NUMBER		20
		DW	0			;HLD			21
		DW	0			;HANDLER		22
		DW	0			;CONTEXT pointer 	23
		DW	0,0,0,0,0,0,0,0	; VOCSS # for vocabulary stack
XCU:		DW	0			;CURRENT pointer	32
XVL:		DW	0			;voc-link pointer	33
XCT:		DW	CTOP			;CP			34	
		DW	NTOP			;NP	NOT USED 2.0	35
XLA:		DW	LASTN			;LAST			36
		DW	0			;			37
		DW	0			;			38
		DW	0			;			39			
		DW	0			;			40
		DW	0			;			41
		DW	0			; KLS added 	42
ULAST:
;;	Be sure that assembly variable US is at least as big!


	ORG	(START_FORTH+US)				;beginning of the code dictionary

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; Device dependent I/O									
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;   BYE		( -- )
;		Exit eForth.  returns to dyn-c.

		M_CODE	3,"BYE",BYE

		ld	sp,(SAVE_SP)	; restore dyn-c sp
					; so we know where we left the regs
		pop	ix
		pop	iy
		pop	af
		pop	hl
		pop	de
		pop	bc		; restore state
		
		ld	H,0		; was baud. now return value.
		ld	L,0
		ret

;;		M_NEXT

;; Trick to keep words out of the vocabulary:
;; Use an assembler variable to save the link to the last word
;; defined in the vocabulary.  The words defined from here until
;; the link variable _LINK is restored point into the main
;; vocabulary, but nothing points to them.  Their addresses are
;; used (referenced by label) in the word list for SREC, which
;; is a linked word.

SAVELINK	EVAL LINK	;; DON'T LINK WORDS USED ONLY BY 
;;				;; WORDS DEFINED HERE IN ASSEMBLER

;   BYE_F	( a -- )
;		Exit eForth.  returns to dyn-c.

		M_CODE	4,"BYEF",BYEF

		pop	HL		; get addr off data stack
					; this is our return value

		ld	sp,(SAVE_SP)	; restore dyn-c sp
					; so we know where we left the regs
		pop	ix
		pop	iy
		pop	af
;		pop	hl		; don't restore over return value
		pop	de		; stack fix up to toss old HL
		pop	de
		pop	bc		; restore state
		
		ret

;;		M_NEXT


LINK	EVAL SAVELINK		; RESTORE LINK



;;  FLASH	( -- )
;;		Exit eForth & flash forth image. 

		M_COLON	5,"flash",FLASH

		; update these user var init values before exiting!
;XCU:		DW	0			;CURRENT pointer	32
;XVL:		DW	0			;voc-link pointer	33
;XCT:		DW	CTOP			;CP			34	
;		DW	NTOP			;NP	NOT USED 2.0	35
;XLA:		DW	LASTN			;LAST			36

; LAST @ ' BYE 2A - !
; CP @ ' BYE 2E - !
		DW	LAST,AT,doLIT,XLA,STORE
		DW	CP,AT,doLIT,XCT,STORE
		DW	HERE
		DW	BYEF
;		M_NEXT


;; FORGET	( -- )
;;		Forgets new words added via FLASH

		M_COLON	6,"forget",FORGET

		DW	doLIT,LASTN,doLIT,XLA,STORE
		DW	doLIT,CTOP,doLIT,XCT,STORE
		DW	doLIT,CTOP		; original value of HERE
		DW	BYEF			; must exit to dyn-c to reflash new values
;		M_NEXT


;   ?RX		( -- c T | F )
;		Return input character and true, or a false if no input.

	M_CODE	3,"?RX",QRX

;hardware specific...
;
; receive byte into reg A
	PIN ((SERIAL+STATREG))        	; get serial port status
	BIT 7,A				; char ready?
	JR Z,NO_RX			; not
	PIN (SERIAL)			; get data
;...
;push char on data stack
	LD	L,A
	LD	H,000H
	PUSH	HL
;push flag on data stack
	LD	L,0FFH
	LD	H,0FFH
	PUSH	HL

	M_NEXT
NO_RX:
; push F flag on data stack
	LD	L,000H
	LD	H,000H
	PUSH	HL

	M_NEXT

;   TX!		( c -- )
;	 	Send character c to the output device.

	M_CODE	3,"TX!",TXSTO

;hardware specific...
;
TXCHAR:
	PIN	((SERIAL+STATREG))     	; get status
	BIT   2,A			; tx busy?
	JR    NZ,TXCHAR			; loop until its not (just serial shift delay)
;get c off data stack (cell size)
	POP	HL			; HL <- (SP)
	LD	A,L			; put c in reg A
	IOI 
	LD	(SERIAL),A    ; send the character
;...
	M_NEXT


;   !IO		( -- )
;		Initialize the serial I/O devices.

	M_CODE	3,"!IO",STOIO

;hardware specific...
; DONE AT POWER UP.  NOTHING TO DO HERE.
;...

	M_NEXT



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; The kernel										
;;											
;; doLIT, doLIST, next, ?branch, branch, EXECUTE, EXIT, !, @, 		
;; P!, P@,	 ( words on this line added by KLS )				
;; RP@, RP!, R>, R@, >R, SP@, SP!, DROP, DUP, SWAP, OVER, 0<, AND, OR,	
;; XOR, UM+								
;; 25 kernel words in EForth model.					
;;									
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;   doLIT	( -- w )
;		Push an inline literal.

		M_CODE	COMPO+5,"doLIT",DOLIT
		LD  A,(BC)	; IP points to literal (inline)
		INC BC    
		LD  L,A   
		LD  A,(BC)
		INC BC    
		LD  H,A   
		PUSH HL		; push it on the stack

		M_NEXT

;   doLIST	( a -- )
;		Process colon list.

		M_CODE	COMPO+6,"doLIST",DOLST1
DOLST:				; KLS MACRO LABEL
					; this is ok in code def...
		LD  HL,(RPr)	; get addr of return stack
		DEC HL     		; decrement it
		LD  (HL),B  	; end of r-push simulation (high byte push)
		DEC HL  		; decrement it again
		LD  (HL),C  	; end of r-push simulation (low byte push)
		LD  (RPr),HL 	; IP is r-pushed (simulated) return ptr points new value

		POP BC      	; get new BC:  BC <- (SP), SP+

		M_NEXT

;   next	( -- )
;		Run time code for the single index loop.
;		: next ( -- ) \ hilevel model
;		  r> r> dup if 1 - >r @ >r exit then drop cell+ >r ;

		M_CODE	COMPO+4,"next",DONXT
		LD  HL,(RPr)
		LD  A,(HL) 
		OR  A      
		JR  NZ,DECLOW 	; a fast dec is ok, only failed every 255 time
                              ; low byte 0
		INC HL    
		LD  A,(HL)
		OR  A       
		JR  NZ,DECHILO 	; Hi-byte no-zero, it is also a re-loop case
                       		;zero bound now .. .
		INC HL       	; discard the loop count on R-stack
		LD  (RPr),HL 
		INC BC       	; IP slip over the re-loop-addr
		INC BC     

		M_NEXT

DECHILO:	DEC (HL)     	; hi-byte
		DEC HL       	; back to low byte
DECLOW:	DEC (HL)     	; low byte non-zero, just dec it and re-loop
		LD  L,C     	; get loop-start-adr to IP and keep stepping
		LD  H,B   
		LD  C,(HL)
		INC HL 
		LD  B,(HL)

		M_NEXT

;   ?branch	( f -- )
;		Branch if flag is zero.

		M_CODE	COMPO+7,"?branch",QBRAN
		POP HL		; HL <-(SP), SP+
		LD  A,L      ; 			?branch adr is just after ?branch
		OR  H         ;  A=A|H    	and IP is pointing it
		JR  Z,ZEROO   ; 
		INC BC        ; IP slip over the retun addr
		INC BC        ; ex: 'TRUE IF' will slip over

		M_NEXT
        
ZEROO:	LD  L,C      ; get IP :=[IP] and go
		LD  H,B 
		LD  C,(HL)
		INC HL     
		LD  B,(HL)

		M_NEXT

;   branch	( -- )
;		Branch to an inline address.

		M_CODE	COMPO+6,"branch",BRAN
		LD  L,C      ;    get IP :=[IP] and go
		LD  H,B     
		LD  C,(HL)  
		INC HL      
		LD  B,(HL)

		M_NEXT

;   EXECUTE	( ca -- )
;		Execute the word at ca.

		M_CODE	7,"EXECUTE",EXECU
		POP HL		; HL <- (SP), SP+
		JP  (HL)
		

;   EXIT	( -- )
;		Terminate a colon definition.

		M_CODE	4,"EXIT",EXIT
		LD  HL,(RPr)
		LD  C,(HL) 
		INC HL     
		LD  B,(HL) 	; BC <- (RPr)
		INC HL 
		LD  (RPr),HL	; RPr+

		M_NEXT

;   !		( w a -- )
;		Pop the data stack to memory.

		M_CODE	1,"!",STORE
		POP HL    	; addr a: HL <- (SP), SP+ 
		POP DE    	; data w: DE <- (SP), SP+
		LD  (HL),E	; low byte data store
		INC HL    
		LD  (HL),D	; high byte data store

		M_NEXT

;   @		( a -- w )
;		Push memory location to the data stack.

		M_CODE	1,"@",AT
		POP HL
		LD  E,(HL)
		INC HL    
		LD  D,(HL)
		PUSH DE   	; data w: (SP-) <- DE

		M_NEXT

;   C!          ( c b -- )
;               Pop the data stack to byte memory.

                M_CODE   2,"C!",CSTOR
		POP HL		; addr b: HL <- (SP), SP+
		POP DE		; data c: DE <- (SP), SP+
		LD  (HL),E 	; low byte data store
      
		M_NEXT


;   C@          ( b -- c )
;               Push byte memory location to the data stack.

            M_CODE   2,"C@",CAT
		POP HL     
		LD  E,(HL)
		LD  D,0
		PUSH DE

		M_NEXT


;   P!          ( c b -- )	**KLS
;               Pop the data stack to internal byte i/o port.

            M_CODE   2,"P!",PIOSTOR
		POP HL		; addr b: HL <- (SP), SP+
		POP DE		; data c: DE <- (SP), SP+
    		IOI 
		LD  (HL),E 	; low byte data store
		M_NEXT


;   P@          ( b -- c )	**KLS
;               Push internal byte i/o port to the data stack.

            M_CODE   2,"P@",PIOAT
		POP HL     
		IOI 
		LD  E,(HL)
		LD  D,0
		PUSH DE
		M_NEXT


;   RP@		( -- a )
;		Push the current RP to the data stack.

		M_CODE	3,"RP@",RPAT
		LD  HL,(RPr)
		PUSH HL
		M_NEXT

;   RP!		( a -- )
;		Set the return stack pointer.

		M_CODE	COMPO+3,"RP!",RPSTO
		POP HL 
		LD  (RPr),HL
		M_NEXT

;   R>		( -- w )
;		Pop the return stack to the data stack.

		M_CODE	COMPO+2,"R>",RFROM
		LD  HL,(RPr)
		LD  E,(HL) 
		INC HL     
		LD  D,(HL) 
		INC HL     
		LD  (RPr),HL
		PUSH DE

		M_NEXT

;   R@		( -- w )
;		Copy top of return stack to the data stack.

		M_CODE	2,"R@",RAT
		LD  HL,(RPr)
		LD  E,(HL) 
		INC HL      
		LD  D,(HL) 
		PUSH DE     

		M_NEXT

;   >R		( w -- )
;		Push the data stack to the return stack.

		M_CODE	COMPO+2,">R",TOR
		POP DE 
		LD  HL,(RPr)
		DEC HL     
		LD  (HL),D 
		DEC HL     
		LD  (HL),E 
		LD  (RPr),HL

		M_NEXT

;   SP@		( -- a )
;		Push the current data stack pointer. (onto the data stack)

		M_CODE	3,"SP@",SPAT
		LD  HL,0
		ADD HL,SP
		PUSH HL
		M_NEXT

;   SP!		( a -- )
;		Set the data stack pointer.

		M_CODE	3,"SP!",SPSTO
		POP HL
		LD  SP,HL
		M_NEXT

;   DROP	( w -- )
;		Discard top stack item.

		M_CODE	4,"DROP",DROP
		POP HL
		M_NEXT

;   DUP		( w -- w w )
;		Duplicate the top stack item.

;		M_CODE	3,"DUP",DUPP
		M_CODE	3,"\x44UP",DUPP
		POP HL 
		PUSH HL
		PUSH HL

		M_NEXT

;   SWAP	( w1 w2 -- w2 w1 )
;		Exchange top two stack items.

		M_CODE	4,"SWAP",SWAP
		POP HL
		DB	0EDH,054H	;EX  (SP),HL	;Rabbit 2000 modified opcode (kls)
		PUSH HL					
		
		M_NEXT

;   OVER	( w1 w2 -- w1 w2 w1 )
;		Copy second stack item to top.

		M_CODE	4,"OVER",OVER
		POP DE 
		POP HL 
		PUSH HL
		PUSH DE
		PUSH HL

		M_NEXT

;   0<		( n -- t )
;		Return true if n is negative.

		M_CODE	2,"0<",ZLESS
		POP HL 
		ADD HL,HL
		JR  C,LESSZ
		LD  HL,0
		PUSH HL

		M_NEXT
                      
LESSZ: 		LD  HL,0FFFFH
		PUSH HL
		
		M_NEXT

;   AND		( w w -- w )
;		Bitwise AND.

		M_CODE	3,"AND",ANDD
		POP DE
		POP HL
		LD  A,E
		AND L
		LD  L,A
		LD  A,D
		AND H
		LD  H,A
		PUSH HL

		M_NEXT

;   OR		( w w -- w )
;		Bitwise inclusive OR.

		M_CODE	2,"OR",ORR
		POP DE
		POP HL
		LD  A,E
		OR  L
		LD  L,A
		LD  A,D
		OR  H
		LD  H,A
		PUSH HL

		M_NEXT

;   XOR		( w w -- w )
;		Bitwise exclusive OR.

		M_CODE	3,"XOR",XORR
		POP DE
		POP HL
		LD  A,E
		XOR L
		LD  L,A
		LD  A,D
		XOR H
		LD  H,A
		PUSH HL
		M_NEXT

;   UM+		( u u -- udsum )
;		Add two unsigned single numbers and return a double sum.

		M_CODE	3,"UM+",UPLUS
		POP DE
		POP HL
		ADD HL,DE
		PUSH HL
		JR  C,CARRY
		LD  HL,0
		PUSH HL

		M_NEXT

CARRY:		LD  HL,1
		PUSH HL

		M_NEXT





;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; REST OF SOURCE IS BASED ON PREVIOUS KERNEL					
;; (except UM/MOD, UM* in code for speed)					
;;										
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;
;; The next three words are needed to define user variable entries
;; in code space 
;; (name space entries made by the assembler macro M_USER).  
;; M_USER refers to doUSER which refers to UP which refers to doVAR.
;;

;   doVAR	( -- a )
;		Run time routine for VARIABLE and CREATE.

		M_COLON	COMPO+5,"doVAR",DOVAR
		DW	RFROM,EXIT

;   UP		( -- a )
;		Pointer to the user area.

		M_COLON	2,"UP",UP
		DW	DOVAR
		DW	UPP

;   doUSER	( -- a )
;		Run time routine for user variables.

;;; FIXME KLS MACRO LABEL
;;  so expand macro by hand here...
;;		M_COLON	COMPO+6,"doUSER",DOUSE1
;;M_COLON	MACRO	LEX,NAMED,LABEL
	align CELLSIZE
	DW	DOUSE,LINK			;;token pointer and link
LINK	EVAL $				;;link points to a name string
	DB	COMPO+6,"doUSER"	;;name string
	align CELLSIZE
DOUSE:					;;assembly label

	DW	CALLL
	DW	DOLST			;;include BSR doLIST
;;	ENDM
		DW	RFROM,AT,UP,AT,PLUS,EXIT

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; User variables									
;; MUST BE IN SAME ORDER AS USER VARIABLES (at UZERO).			
;;									
;; These all point to entries in the user area in RAM who's values	
;; are copied from ROM at COLD start.					
;;									
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;   SP0		( -- a )
;		Pointer to bottom of the data stack.

		M_USER	3,"SP0",SZERO

;   RP0		( -- a )
;		Pointer to bottom of the return stack.

		M_USER	3,"RP0",RZERO

;   '?KEY	( -- a )
;		Execution vector of ?KEY.

		M_USER	5,"'?KEY",TQKEY

;   'EMIT	( -- a )
;		Execution vector of EMIT.

		M_USER	5,"'EMIT",TEMIT

;   'EXPECT	( -- a )
;		Execution vector of EXPECT.

		M_USER	7,"'EXPECT",TEXPE

;   'TAP	( -- a )
;		Execution vector of TAP.

		M_USER	4,"'TAP",TTAP

;   'ECHO	( -- a )
;		Execution vector of ECHO.

		M_USER	5,"'ECHO",TECHO

;   'PROMPT	( -- a )
;		Execution vector of PROMPT.

		M_USER	7,"'PROMPT",TPROM

;   BASE	( -- a )
;		Storage of the radix base for numeric I/O.

		M_USER	4,"BASE",BASE

;   tmp		( -- a )
;		A temporary storage location used in parse and find.

		M_USER	COMPO+3,"tmp",TEMP

;   SPAN	( -- a )
;		Hold character count received by EXPECT.

		M_USER	4,"SPAN",SPAN

;   >IN		( -- a )
;		Hold the character pointer while parsing input stream.

		M_USER	3,">IN",INN

;   #TIB	( -- a )
;		Hold the current count in and address of the terminal input buffer.

		M_USER	4,"#TIB",NTIB	;hold the count of characters in buffer
USERV 	EVAL	USERV+CELLL		;hold the base address of the terminal input buffer

;   CSP		( -- a )
;		Hold the stack pointer for error checking.

		M_USER	3,"CSP",CSP

;   'EVAL	( -- a )
;		Execution vector of EVAL.

		M_USER	5,"'EVAL",TEVAL

;   'NUMBER	( -- a )
;		Execution vector of NUMBER?.

		M_USER	7,"'NUMBER",TNUMB

;   HLD		( -- a )
;		Hold a pointer in building a numeric output string.

		M_USER	3,"HLD",HLD

;   HANDLER	( -- a )
;		Hold the return stack pointer for error handling.

		M_USER	7,"HANDLER",HANDL

;   CONTEXT	( -- a )
;		A area to specify vocabulary search order.
;		This is a stack of vocabulary variable addresses.

		M_USER	7,"CONTEXT",CNTXT

USERV	EVAL	USERV+VOCSS*CELLL	;vocabulary stack


;   CURRENT	( -- a )
;		Point to the vocabulary to be extended.

		M_USER	7,"CURRENT",CRRNT	;vocabulary name pointer

;   voc-link	( -- a )
;		Point to the most recently defined vocabulary.

		M_USER	8,"voc-link",VOCLINK	;vocabulary link pointer

;   CP		( -- a )
;		Point to the top of the code dictionary.

		M_USER	2,"CP",CP

;   NP		( -- a )
;		Point to the bottom of the name dictionary.

		M_USER	2,"NP",NP

;   LAST	( -- a )
;		Point to the last name in the name dictionary.

		M_USER	4,"LAST",LAST


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; Common functions									
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;   doDOES	( -- )		(**KLS**)
;		Run time code of DOES>
;
		M_COLON	COMPO+6,"doDOES",doDOES
		DW	LAST,AT			; last defined word's link (via variable)
		DW	CELLM,CELLM,AT		; address of word's code (via code pointer)
		DW	CELLP,CELLP		; skip CALLN & DOLST  ***CHECK THIS***
		DW	RFROM			; get instr ptr off return stack
		DW	CELLP			; skip COMPILE
		DW	SWAP,STORE,EXIT		; write address of 1st list item over EXIT


;;   doVOC	( -- )
;;		Run time action of VOCABULARY. SEE VOCABULARY DEFINITION...
;;
;;		M_COLON	COMPO+5,"doVOC",DOVOC
;;		DW	RFROM,CNTXT,STORE,EXIT



;   FORTH	( -- )		(**KLS**)
;		Make FORTH the context vocabulary.
;		Simulate run time portion of VOCABULARY
;
; need to fully expand m_colon macro here to include voc-link ,
; and make cfa point to runtime of vocabulary!
;;  so expand macro by hand here...
;;		M_COLON	5,"FORTH",FORTH
;;M_COLON	MACRO	LEX,NAMED,LABEL
	align CELLSIZE
	DW	FORTH,LINK			;;token pointer and link
LINK	EVAL $				;;link points to a name string
	DB	5,"FORTH"		;;name string
	align CELLSIZE
FORTH:					;;assembly label

	DW	CALLL
	DW	DOLST			;;include BSR doLIST
;;	ENDM

	DW	doVOC

	DW	LASTN			; NAME link
	DW	0			; NULL voc-link




;   ?DUP	( w -- w w | 0 )
;		Dup tos if its is not zero.

		M_COLON	4,"?\x44UP",QDUP
		DW	DUPP
		DW	QBRAN,QDUP1
		DW	DUPP
QDUP1:		DW	EXIT

;   ROT		( w1 w2 w3 -- w2 w3 w1 )
;		Rot 3rd item to top.

		M_COLON	3,"ROT",ROT
		DW	TOR,SWAP,RFROM,SWAP,EXIT

;   2DROP	( w w -- )
;		Discard two items on stack.

		M_COLON	5,"2DROP",DDROP
		DW	DROP,DROP,EXIT

;   2DUP	( w1 w2 -- w1 w2 w1 w2 )
;		Duplicate top two items.

		M_COLON	4,"2DUP",DDUP
		DW	OVER,OVER,EXIT

;   +		( w w -- sum )
;		Add top two items.

		M_COLON	1,"+",PLUS
		DW	UPLUS,DROP,EXIT

;   NOT		( w -- w )
;		One's complement of tos.

		M_COLON	3,"NOT",INVER
		DW	DOLIT,-1,XORR,EXIT

;   NEGATE	( n -- -n )
;		Two's complement of tos.

		M_COLON	6,"NEGATE",NEGAT
		DW	INVER,DOLIT,1,PLUS,EXIT

;   DNEGATE	( d -- -d )
;		Two's complement of top double.

		M_COLON	7,"DNEGATE",DNEGA
		DW	INVER,TOR,INVER
		DW	DOLIT,1,UPLUS
		DW	RFROM,PLUS,EXIT

;   -		( n1 n2 -- n1-n2 )
;		Subtraction.

		M_COLON	1,"-",SUBB
		DW	NEGAT,PLUS,EXIT

;   ABS		( n -- n )
;		Return the absolute value of n.

		M_COLON	3,"ABS",ABSS
		DW	DUPP,ZLESS
		DW	QBRAN,ABS1
		DW	NEGAT
ABS1:		DW	EXIT

;   =		( w w -- t )
;		Return true if top two are equal.

		M_COLON	1,"=",EQUAL
		DW	XORR
		DW	QBRAN,EQU1
		DW	DOLIT,0,EXIT
EQU1:		DW	DOLIT,TRUEE,EXIT

;   U<		( u u -- t )
;		Unsigned compare of top two items.

		M_COLON	2,"U<",ULESS
		DW	DDUP,XORR,ZLESS
		DW	QBRAN,ULES1
		DW	SWAP,DROP,ZLESS,EXIT
ULES1:		DW	SUBB,ZLESS,EXIT

;   <		( n1 n2 -- t )
;		Signed compare of top two items.

		M_COLON	1,"<",LESS
		DW	DDUP,XORR,ZLESS
		DW	QBRAN,LESS1
		DW	DROP,ZLESS,EXIT
LESS1:		DW	SUBB,ZLESS,EXIT

;   MAX		( n n -- n )
;		Return the greater of two top stack items.

		M_COLON	3,"MAX",MAX
		DW	DDUP,LESS
		DW	QBRAN,MAX1
		DW	SWAP
MAX1:		DW	DROP,EXIT

;   MIN		( n n -- n )
;		Return the smaller of top two stack items.

		M_COLON	3,"MIN",MIN
		DW	DDUP,SWAP,LESS
		DW	QBRAN,MIN1
		DW	SWAP
MIN1:		DW	DROP,EXIT

;   WITHIN	( u ul uh -- t )
;		Return true if u is within the range of ul and uh. ( ul <= u < uh )

		M_COLON	6,"WITHIN",WITHI
		DW	OVER,SUBB,TOR
		DW	SUBB,RFROM,ULESS,EXIT

;; Divide

;   UM/MOD	( udl udh un -- ur uq )
;		Unsigned divide of a double by a single. Return mod and quotient.

;;( shift dividend left method )
;;
;; AQ dividend 
;; B divisor
;; left shift AQ n times
;; Q quotient. A remainder.
;;
;; A A UM+ 2AL 2AH   ( DROP 2AH 0/1 ONLY )
;; Q Q UM+ 2QL 2QH   ( 2QH = 0 or 1 ONLY )
;; AQ' = (2AL + 2QH) 2QL fully shifted
;; 
;;2DUP ( Q A u A u -- )
;;U<  ( Q A u t|f -- )
;;IF NEGATE ( Q A -u -- )
;;  32 ( SET TO NUMBER OF BITS IN WORD SIZE )
;;  BEGIN DUP ( Q A -u 15 15 -- )
;;  WHILE 
;;    >R >R ( push count & divisor on ret stack )
;;    ( Q A -- )
;;    DUP ( Q A A -- ) UM+ ( double A = 2AL 2AH )
;;    >R >R ( push doubled A on ret stack )
;;    DUP ( Q Q -- ) UM+ ( double Q = 2QL 2QH )
;;    R> ( pop low word of dbl A ) ( 2QL 2QH 2AL -- )
;;    + ( Q' A' -- )
;;    DUP ( Q' A' A' -- )
;;    R> ( pop high word of dbl A ) ( Q' A' A' 2AH -- )
;;    R@ ( _copy_ of divisor from ret stack ) 
;;    SWAP ( Q' A' A' -u 2AH --- )
;;    >R ( push 2AH )
;;    UM+ ( subtract divisor from new A, DL DH )
;;    R> ( pop 2AH ) ( Q' A' DL DH 2AH -- )
;;    OR ( DH | 2AH ) ( Q' A' DL t|f -- )
;;    ( logical result determines return stack action )
;;    IF 
;;      >R ( push A' ) 
;;      DROP ( DL ) 
;;      1 + ( set Qn bit ) 
;;      R> ( restore A' ) 
;;    ELSE DROP ( DL )
;;    THEN 
;;    R> R> ( retrieve divisor, then count )
;;    1 - ( decrement count )
;;  REPEAT
;;2DROP ( drop divisor, count ) SWAP EXIT ( ur uq -- )
;;THEN 
;;( if here, then dividend > 2x divisor )
;;DROP 2DROP -1 DUP ; ( -1 -1 -- )

;
; 2DUP U<
; IF NEGATE 15 ( 16 bit forth, n+1 FOR-NEXT iterations )
;   FOR >R DUP UM+ >R >R DUP UM+ R> + DUP
;       R> R@ SWAP >R UM+ R> OR
;     IF >R DROP 1 + R> ELSE DROP THEN R>
;   NEXT DROP SWAP EXIT
; THEN DROP 2DROP -1 DUP ;
;


		M_COLON	6,"UM/MOD",UMMOD

                DW      DDUP,ULESS
                DW      QBRAN,UMM4
                DW      NEGAT,DOLIT,15,TOR
UMM1:		DW      TOR,DUPP,UPLUS
                DW      TOR,TOR,DUPP,UPLUS
                DW      RFROM,PLUS,DUPP
                DW      RFROM,RAT,SWAP,TOR
                DW      UPLUS,RFROM,ORR
                DW      QBRAN,UMM2
                DW      TOR,DROP,DOLIT,1,PLUS,RFROM
                DW      BRAN,UMM3
UMM2:           DW      DROP
UMM3:           DW      RFROM
                DW      DONXT,UMM1
                DW      DROP,SWAP,EXIT
UMM4:           DW      DROP,DDROP
                DW      DOLIT,-1,DUPP,EXIT      ;overflow, return max



;   M/MOD	( d n -- r q )
;		Signed floored divide of double by single. Return mod and quotient.

		M_COLON	5,"M/MOD",MSMOD
		DW	DUPP,ZLESS,DUPP,TOR
		DW	QBRAN,MMOD1
		DW	NEGAT,TOR,DNEGA,RFROM
MMOD1:		DW	TOR,DUPP,ZLESS
		DW	QBRAN,MMOD2
		DW	RAT,PLUS
MMOD2:		DW	RFROM,UMMOD,RFROM
		DW	QBRAN,MMOD3
		DW	SWAP,NEGAT,SWAP
MMOD3:		DW	EXIT

;   /MOD	( n n -- r q )
;		Signed divide. Return mod and quotient.

		M_COLON	4,"/MOD",SLMOD
		DW	OVER,ZLESS,SWAP,MSMOD,EXIT

;   MOD		( n n -- r )
;		Signed divide. Return mod only.

		M_COLON	3,"MOD",MODD
		DW	SLMOD,DROP,EXIT

;   /		( n n -- q )
;		Signed divide. Return quotient only.

		M_COLON	1,"/",SLASH
		DW	SLMOD,SWAP,DROP,EXIT

;; Multiply

;   UM*		( u1 u2 -- ud )
;		Unsigned multiply. Return double product.

		M_COLON	3,"UM*",UMSTA

                DW      DOLIT,0,SWAP,DOLIT,15,TOR
UMST1:          DW      DUPP,UPLUS,TOR,TOR
                DW      DUPP,UPLUS,RFROM,PLUS,RFROM
                DW      QBRAN,UMST2
                DW      TOR,OVER,UPLUS,RFROM,PLUS
UMST2:          DW      DONXT,UMST1
                DW      ROT,DROP,EXIT


;   *		( n n -- n )
;		Signed multiply. Return single product.

		M_COLON	1,"*",STAR
		DW	UMSTA,DROP,EXIT

;   M*		( n n -- d )
;		Signed multiply. Return double product.

		M_COLON	2,"M*",MSTAR
		DW	DDUP,XORR,ZLESS,TOR
		DW	ABSS,SWAP,ABSS,UMSTA
		DW	RFROM
		DW	QBRAN,MSTA1
		DW	DNEGA
MSTA1:		DW	EXIT

;   */MOD	( n1 n2 n3 -- r q )
;		Multiply n1 and n2, then divide by n3. Return mod and quotient.

		M_COLON	5,"*/MOD",SSMOD
		DW	TOR,MSTAR,RFROM,MSMOD,EXIT

;   */		( n1 n2 n3 -- q )
;		Multiply n1 by n2, then divide by n3. Return quotient only.

		M_COLON	2,"*/",STASL
		DW	SSMOD,SWAP,DROP,EXIT

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; Miscellaneous									
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;   CELL+	( a -- a )
;		Add cell size in byte to address.

		M_COLON	5,"CELL+",CELLP
		DW	DOLIT
		DW	CELLL
		DW	PLUS,EXIT

;   CELL-	( a -- a )
;		Subtract cell size in byte from address.

		M_COLON	5,"CELL-",CELLM
		DW	DOLIT
;kls;		DW	-1		;intend sign extended -4.L
;kls;		DB	-1
;kls;		DB	-4		;DW  0-CELLL
		DW	-CELLL
		DW	PLUS,EXIT

;   CELLS	( n -- n )
;		Multiply tos by cell size in bytes.

		M_COLON	5,"CELLS",CELLS
		DW	DOLIT,CELLL,STAR,EXIT

;   ALIGNED	( b -- a )
;		Align address to the cell boundary.
;   		note	UM/MOD ( udl udh un -- ur uq )

		M_COLON	7,"ALIGNED",ALGND
		DW	DUPP,DOLIT,0,DOLIT,CELLL	;dbl b, div by celll
		DW	UMMOD,DROP,DUPP			;lose quotient
		DW	QBRAN,ALGN1			;rem=0, goto algn1
		DW	DOLIT,CELLL,SWAP,SUBB		;celll-rem=adjust
ALGN1:		DW	PLUS,EXIT			;a=b+0 or a=b+adjust

;   BL		( -- 32 )
;		Return 32, the blank character.

		M_COLON	2,"BL",BLANK
		DW	DOLIT,' ',EXIT

;   >CHAR	( c -- c )
;		Filter non-printing characters.

		M_COLON	5,">CHAR",TCHAR
		DW	DOLIT,07FH,ANDD,DUPP	;mask msb
		DW	DOLIT,127,BLANK,WITHI	;check for printable
		DW	QBRAN,TCHA1
		DW	DROP,DOLIT,'_'		;replace non-printables
TCHA1:		DW	EXIT

;   DEPTH	( -- n )
;		Return the depth of the data stack.

		M_COLON	5,"DEPTH",DEPTH
		DW	SPAT,SZERO,AT,SWAP,SUBB
		DW	DOLIT,CELLL,SLASH,EXIT

;   PICK	( ... +n -- ... w )
;		Copy the nth stack item to tos.

		M_COLON	4,"PICK",PICK
		DW	DOLIT,1,PLUS,CELLS
		DW	SPAT,PLUS,AT,EXIT

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; Memory access									
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;   +!		( n a -- )
;		Add n to the contents at address a.

		M_COLON	2,"+!",PSTOR
		DW	SWAP,OVER,AT,PLUS
		DW	SWAP,STORE,EXIT

;   2!		( d a -- )
;		Store the double integer to address a.

		M_COLON	2,"2!",DSTOR
		DW	SWAP,OVER,STORE
		DW	CELLP,STORE,EXIT

;   2@		( a -- d )
;		Fetch double integer from address a.

		M_COLON	2,"2@",DAT
		DW	DUPP,CELLP,AT
		DW	SWAP,AT,EXIT

;   COUNT	( b -- b +n )
;		Return count byte of a string and add 1 to byte address.

		M_COLON	5,"COUNT",COUNT
		DW	DUPP,DOLIT,1,PLUS
		DW	SWAP,CAT,EXIT

;   HERE	( -- a )
;		Return the top of the code dictionary.

		M_COLON	4,"HERE",HERE
		DW	CP,AT,EXIT

;   PAD		( -- a )
;		Return the address of the text buffer above the code dictionary.

		M_COLON	3,"PAD",PAD
		DW	HERE,DOLIT,80,PLUS,EXIT

;   TIB		( -- a )
;		Return the address of the terminal input buffer.

		M_COLON	3,"TIB",TIB
		DW	NTIB,CELLP,AT,EXIT

;   @EXECUTE	( a -- )
;		Execute vector stored in address a.

		M_COLON	8,"@EXECUTE",ATEXE
		DW	AT,QDUP			;?address or zero
		DW	QBRAN,EXE1
		DW	EXECU			;execute if non-zero
EXE1:		DW	EXIT			;do nothing if zero

;   CMOVE	( b1 b2 u -- )
;		Copy u bytes from b1 to b2.

		M_COLON	5,"CMOVE",CMOVE
		DW	TOR
		DW	BRAN,CMOV2
CMOV1:		DW	TOR,DUPP,CAT
		DW	RAT,CSTOR
		DW	DOLIT,1,PLUS
		DW	RFROM,DOLIT,1,PLUS
CMOV2:		DW	DONXT,CMOV1
		DW	DDROP,EXIT

;   FILL	( b u c -- )
;		Fill u bytes of character c to area beginning at b.

		M_COLON	4,"FILL",FILL
		DW	SWAP,TOR,SWAP
		DW	BRAN,FILL2
FILL1:		DW	DDUP,CSTOR,DOLIT,1,PLUS
FILL2:		DW	DONXT,FILL1
		DW	DDROP,EXIT

;   -TRAILING	( b u -- b u )
;		Adjust the count to eliminate trailing white space.

		M_COLON	9,"-TRAILING",DTRAI
		DW	TOR
		DW	BRAN,DTRA2
DTRA1:		DW	BLANK,OVER,RAT,PLUS,CAT,LESS
		DW	QBRAN,DTRA2
		DW	RFROM,DOLIT,1,PLUS,EXIT
DTRA2:		DW	DONXT,DTRA1
		DW	DOLIT,0,EXIT

;   PACK$	( b u a -- a )
;		Build a counted string with u characters from b. Null fill.

		M_COLON	5,"PACK$",PACKS
		DW	ALGND,DUPP,TOR		;strings only on cell boundary
		DW	OVER,DUPP,DOLIT,0
		DW	DOLIT,CELLL,UMMOD,DROP	;count mod cell
		DW	SUBB,OVER,PLUS
		DW	DOLIT,0,SWAP,STORE	;null fill cell
		DW	DDUP,CSTOR,DOLIT,1,PLUS	;save count
		DW	SWAP,CMOVE,RFROM,EXIT	;move string

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;										
;; Numeric output, single precision						
;;										
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;   DIGIT	( u -- c )
;		Convert digit u to a character.

		M_COLON	5,"DIGIT",DIGIT
		DW	DOLIT,9,OVER,LESS	;(9 u -- ) less
		DW	DOLIT,7,ANDD,PLUS	; 0111
		DW	DOLIT,'0',PLUS,EXIT

;   EXTRACT	( n base -- n c )
;		Extract the least significant digit from n.

		M_COLON	7,"EXTRACT",EXTRC
		DW	DOLIT,0,SWAP,UMMOD	; ( n 0 base -- ) divide
		DW	SWAP,DIGIT,EXIT

;   <#		( -- )
;		Initiate the numeric output process.

		M_COLON	2,"<#",BDIGS
		DW	PAD,HLD,STORE,EXIT

;   HOLD	( c -- )
;		Insert a character into the numeric output string.

		M_COLON	4,"HOLD",HOLD
		DW	HLD,AT,DOLIT,1,SUBB
		DW	DUPP,HLD,STORE,CSTOR,EXIT

;   #		( u -- u )
;		Extract one digit from u and append the digit to output string.

		M_COLON	1,"#",DIG
		DW	BASE,AT,EXTRC,HOLD,EXIT

;   #S		( u -- 0 )
;		Convert u until all digits are added to the output string.

		M_COLON	2,"#S",DIGS
DIGS1:		DW	DIG,DUPP
		DW	QBRAN,DIGS2
		DW	BRAN,DIGS1
DIGS2:		DW	EXIT

;   SIGN	( n -- )
;		Add a minus sign to the numeric output string.

		M_COLON	4,"SIGN",SIGN
		DW	ZLESS
		DW	QBRAN,SIGN1
		DW	DOLIT,'-',HOLD
SIGN1:		DW	EXIT

;   #>		( w -- b u )
;		Prepare the output string to be TYPE'd.

		M_COLON	2,"#>",EDIGS
		DW	DROP,HLD,AT
		DW	PAD,OVER,SUBB,EXIT

;   str		( w -- b u )
;		Convert a signed integer to a numeric string.

		M_COLON	3,"str",STR
		DW	DUPP,TOR,ABSS
		DW	BDIGS,DIGS,RFROM
		DW	SIGN,EDIGS,EXIT

;   HEX		( -- )
;		Use radix 16 as base for numeric conversions.

		M_COLON	3,"HEX",HEX
		DW	DOLIT,16,BASE,STORE,EXIT

;   DECIMAL	( -- )
;		Use radix 10 as base for numeric conversions.

		M_COLON	7,"DECIMAL",DECIM
		DW	DOLIT,10,BASE,STORE,EXIT

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;										
;; Numeric input, single precision						
;;										
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;   DIGIT?	( c base -- u t )
;		Convert a character to its numeric value. A flag indicates success.

		M_COLON	6,"DIGIT?",DIGTQ
		DW	TOR,DOLIT,'0',SUBB
		DW	DOLIT,9,OVER,LESS
		DW	QBRAN,DGTQ1
		DW	DOLIT,7,SUBB
		DW	DUPP,DOLIT,10,LESS,ORR
DGTQ1:		DW	DUPP,RFROM,ULESS,EXIT

;   NUMBER?	( a -- n T | a F )
;		Convert a number string to integer. Push a flag on tos.

		M_COLON	7,"NUMBER?",NUMBQ
		DW	BASE,AT,TOR,DOLIT,0,OVER,COUNT
		DW	OVER,CAT,DOLIT,'$',EQUAL
		DW	QBRAN,NUMQ1
		DW	HEX,SWAP,DOLIT,1,PLUS
		DW	SWAP,DOLIT,1,SUBB
NUMQ1:		DW	OVER,CAT,DOLIT,'-',EQUAL,TOR
		DW	SWAP,RAT,SUBB,SWAP,RAT,PLUS,QDUP
		DW	QBRAN,NUMQ6
		DW	DOLIT,1,SUBB,TOR
NUMQ2:		DW	DUPP,TOR,CAT,BASE,AT,DIGTQ
		DW	QBRAN,NUMQ4
		DW	SWAP,BASE,AT,STAR,PLUS,RFROM
		DW	DOLIT,1,PLUS
		DW	DONXT,NUMQ2
		DW	RAT,SWAP,DROP
		DW	QBRAN,NUMQ3
		DW	NEGAT
NUMQ3:		DW	SWAP
		DW	BRAN,NUMQ5
NUMQ4:		DW	RFROM,RFROM,DDROP,DDROP,DOLIT,0
NUMQ5:		DW	DUPP
NUMQ6:		DW	RFROM,DDROP
		DW	RFROM,BASE,STORE
		DW	EXIT

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;										
;; Basic I/O									
;;										
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;   ?KEY	( -- c T | F )
;		Return input character and true, or a false if no input.

		M_COLON	4,"?KEY",QKEY
		DW	TQKEY,ATEXE,EXIT

;   KEY		( -- c )
;		Wait for and return an input character.

		M_COLON	3,"KEY",KEY
KEY1:		DW	QKEY
		DW	QBRAN,KEY1
		DW	EXIT

;   EMIT	( c -- )
;		Send a character to the output device.

		M_COLON	4,"EMIT",EMIT
		DW	TEMIT,ATEXE,EXIT

;   NUF?	( -- t )
;		Return false if no input, else pause and if CR return true.

		M_COLON	4,"NUF?",NUFQ
		DW	QKEY,DUPP
		DW	QBRAN,NUFQ1
		DW	DDROP,KEY,DOLIT,CRR,EQUAL
NUFQ1:		DW	EXIT

;   PACE	( -- )
;		Send a pace character for the file downloading process.

		M_COLON	4,"PACE",PACE
;		DW	DOLIT,11H,EMIT,EXIT	; ^Q = 11H (XON)
		DW	DOLIT,2EH,EMIT,EXIT	; use "." for line flow control

;   SPACE	( -- )
;		Send the blank character to the output device.

		M_COLON	5,"SPACE",SPACE
		DW	BLANK,EMIT,EXIT

;   SPACES	( +n -- )
;		Send n spaces to the output device.

		M_COLON	6,"SPACES",SPACS
		DW	DOLIT,0,MAX,TOR
		DW	BRAN,CHAR2
CHAR1:		DW	SPACE
CHAR2:		DW	DONXT,CHAR1
		DW	EXIT

;   TYPE	( b u -- )
;		Output u characters from b.

		M_COLON	4,"TYPE",TYPES
		DW	TOR
		DW	BRAN,TYPE2
TYPE1:		DW	DUPP,CAT,EMIT
		DW	DOLIT,1,PLUS
TYPE2:		DW	DONXT,TYPE1
		DW	DROP,EXIT

;   CR		( -- )
;		Output a carriage return and a line feed.

		M_COLON	2,"CR",CR
		DW	DOLIT,CRR,EMIT
		DW	DOLIT,LF,EMIT,EXIT

;   do$		( -- a )
;		Return the address of a compiled string.

		M_COLON	COMPO+3,"do$",DOSTR
		DW	RFROM,RAT,RFROM,COUNT,PLUS	; R> R@ R> COUNT +
		DW	ALGND,TOR,SWAP,TOR,EXIT

;   $"|		( -- a )
;		Run time routine compiled by $". Return address of a compiled string.

;kls;		M_COLON	COMPO+3,"$"|",STRQP
		M_COLON	COMPO+3,"$\x22|",STRQP
		DW	DOSTR,EXIT		;force a call to do$


;   ."|		( -- )
;		Run time routine of ." . Output a compiled string.
;		MUST be followed by a number of bytes divisible
;		by the cell size (including the count byte)!!

		M_COLON	COMPO+3,".\x22|",DOTQP
		DW	DOSTR,COUNT,TYPES,EXIT

;   .R		( n +n -- )
;		Display an integer in a field of n columns, right justified.

		M_COLON	2,".R",DOTR
		DW	TOR,STR,RFROM,OVER,SUBB
		DW	SPACS,TYPES,EXIT

;   U.R		( u +n -- )
;		Display an unsigned integer in n column, right justified.

		M_COLON	3,"U.R",UDOTR
		DW	TOR,BDIGS,DIGS,EDIGS
		DW	RFROM,OVER,SUBB
		DW	SPACS,TYPES,EXIT

;   U.R2	( u +n -- )		( KLS_WORD )
;		Display a 2 digit unsigned integer in n column, right justified.

		M_COLON	4,"U.R2",UDOTR2
		DW	TOR,BDIGS,DIG,DIG,EDIGS
		DW	RFROM,OVER,SUBB
		DW	SPACS,TYPES,EXIT

;   U.		( u -- )
;		Display an unsigned integer in free format.

		M_COLON	2,"U.",UDOT
		DW	BDIGS,DIGS,EDIGS
		DW	SPACE,TYPES,EXIT

;   .		( w -- )
;		Display an integer in free format, preceeded by a space.

		M_COLON	1,".",DOT
		DW	BASE,AT,DOLIT,10,XORR	;?decimal
		DW	QBRAN,DOT1
		DW	UDOT,EXIT		;no, display unsigned
DOT1:		DW	STR,SPACE,TYPES,EXIT	;yes, display signed

;   ?		( a -- )
;		Display the contents in a memory cell.

		M_COLON	1,"?",QUEST
		DW	AT,DOT,EXIT

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; Parsing										
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;   parse	( b u c -- b u delta ; <string> )
;		Scan string delimited by c. Return found string and its offset.

		M_COLON	5,"parse",PARS
		DW	TEMP,STORE,OVER,TOR,DUPP
		DW	QBRAN,PARS8
		DW	DOLIT,1,SUBB,TEMP,AT,BLANK,EQUAL
		DW	QBRAN,PARS3
		DW	TOR
PARS1:		DW	BLANK,OVER,CAT		;skip leading blanks ONLY
		DW	SUBB,ZLESS,INVER
		DW	QBRAN,PARS2
		DW	DOLIT,1,PLUS
		DW	DONXT,PARS1
		DW	RFROM,DROP,DOLIT,0,DUPP,EXIT
PARS2:		DW	RFROM
PARS3:		DW	OVER,SWAP
		DW	TOR
PARS4:		DW	TEMP,AT,OVER,CAT,SUBB	;scan for delimiter
		DW	TEMP,AT,BLANK,EQUAL
		DW	QBRAN,PARS5
		DW	ZLESS
PARS5:		DW	QBRAN,PARS6
		DW	DOLIT,1,PLUS
		DW	DONXT,PARS4
		DW	DUPP,TOR
		DW	BRAN,PARS7
PARS6:		DW	RFROM,DROP,DUPP
		DW	DOLIT,1,PLUS,TOR
PARS7:		DW	OVER,SUBB
		DW	RFROM,RFROM,SUBB,EXIT
PARS8:		DW	OVER,RFROM,SUBB,EXIT

;   PARSE	( c -- b u ; <string> )
;		Scan input stream and return counted string delimited by c.

		M_COLON	5,"PARSE",PARSE
		DW	TOR,TIB,INN,AT,PLUS	;current input buffer pointer
		DW	NTIB,AT,INN,AT,SUBB	;remaining count
		DW	RFROM,PARS,INN,PSTOR,EXIT

;   .(		( -- )
;		Output following string up to next ) .

		M_COLON	IMEDD+2,".(",DOTPR
		DW	DOLIT,')',PARSE,TYPES,EXIT

;   (		( -- )
;		Ignore following string up to next ) . A comment.

		M_COLON	IMEDD+1,"(",PAREN
		DW	DOLIT,')',PARSE,DDROP,EXIT

;   \		( -- )
;		Ignore following text till the end of line.

		M_COLON	IMEDD+1,"\x5C",BKSLA
		DW	NTIB,AT,INN,STORE,EXIT

;   CHAR	( -- c )
;		Parse next word and return its first character.

		M_COLON	4,"CHAR",CHAR
		DW	BLANK,PARSE,DROP,CAT,EXIT

;   TOKEN	( -- a ; <string> )
;		Parse a word from input stream and copy it to name dictionary.

		M_COLON	5,"TOKEN",TOKEN
		DW	BLANK,PARSE,DOLIT,31,MIN	; actual length or 31
;; v1.x		DW	NP,AT,OVER,SUBB,CELLM
		DW	CP,AT				; HERE
		DW	ALGND				; be sure cell aligned
		DW	CELLP				; leave space for code pointer CFA
		DW	CELLP				; leave space for link pointer LFA
		DW	PACKS				; packed string NFA
		DW	EXIT

;   WORD	( c -- a ; <string> )
;		Parse a word from input stream and copy it to code dictionary.

		M_COLON	4,"WORD",WORDD
		DW	PARSE,HERE,PACKS,EXIT

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;										
;; Dictionary search								
;;										
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;   NAME>	( na -- ca )
;		Return a code address given a name address.

		M_COLON	5,"NAME>",NAMET
		DW	CELLM,CELLM,AT,EXIT 	;( na-8 @ )

;   SAME?	( a a u -- a a f \ -0+ )
;		Compare u cells in two strings. Return 0 if identical.

		M_COLON	5,"SAME?",SAMEQ
		DW	TOR
		DW	BRAN,SAME2
SAME1:		DW	OVER,RAT,CELLS,PLUS,AT
		DW	OVER,RAT,CELLS,PLUS,AT
		DW	SUBB,QDUP
		DW	QBRAN,SAME2
		DW	RFROM,DROP,EXIT		; not same f=difference
SAME2:		DW	DONXT,SAME1
		DW	DOLIT,0,EXIT		; same

;   find	( a va -- ca na | a F )
;		Search a vocabulary for a string. Return ca and na if succeeded.
;
; :find
; SWAP 			\ va a
; DUP C@ 2 / temp ! 	\ va a       get cell count
; DUP @ >R		\ va a
; CELL+ SWAP		\ a' va
; BEGIN @ DUP		\ a' na na
;   IF DUP @ [ =MASK ] LITERAL AND R@ XOR	\ ignore lexicon bits
;     IF CELL+ -1 ELSE CELL+ temp @ SAME? THEN
;   ELSE R> DROP EXIT
;   THEN
; WHILE 2 CELLS - 	\ a' la
; REPEAT R> DROP NIP 1 CELLS - DUP NAME> SWAP ;
;
		M_COLON	4,"find",FIND
		DW	SWAP,DUPP,CAT			;get length in chars
		DW	DOLIT,CELLL,SLASH,TEMP,STORE	;temp=length/celll (# cells)
		DW	DUPP,AT,TOR,CELLP,SWAP		; a' va
FIND1:		DW	AT,DUPP				; a' na na (BEGIN)
		DW	QBRAN,FIND6			; IF...
		DW	DUPP,AT				; get string
		DW	DOLIT,MASKK			; 
		DW	ANDD,RAT,XORR			;
		DW	QBRAN,FIND2
		DW	CELLP,DOLIT,-1			; lex bits 
		DW	BRAN,FIND3
FIND2:		DW	CELLP,TEMP,AT,SAMEQ		;compare strings
FIND3:		DW	BRAN,FIND4
FIND6:		DW	RFROM,DROP			;ELSE clean up R stack
		DW	SWAP,CELLM,SWAP,EXIT		;not found 
FIND4:		DW	QBRAN,FIND5			;THEN
		DW	CELLM,CELLM			;(WHILE) link address
		DW	BRAN,FIND1
FIND5:		DW	RFROM,DROP,SWAP,DROP		;clean up R stack
		DW	CELLM
		DW	DUPP,NAMET,SWAP,EXIT		;found - get ca

;   NAME?	( a -- ca na | a F )
;		Search all context vocabularies for a string.

		M_COLON	5,"NAME?",NAMEQ
		DW	CNTXT,DUPP,DAT,XORR	;context <> also
		DW	QBRAN,NAMQ1
		DW	CELLM			;no, start with context
NAMQ1:		DW	TOR
NAMQ2:		DW	RFROM,CELLP,DUPP,TOR	;next in search order
		DW	AT,QDUP
		DW	QBRAN,NAMQ3
		DW	FIND,QDUP		;search vocabulary
		DW	QBRAN,NAMQ2
		DW	RFROM,DROP,EXIT		;found name exit
NAMQ3:		DW	RFROM,DROP		;name not found
		DW	DOLIT,0,EXIT		;false flag exit

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; Terminal response									
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;   ^H		( bot eot cur -- bot eot cur )
;		Backup the cursor by one character.

		M_COLON	2,"^H",BKSP
		DW	TOR,OVER,RFROM,SWAP,OVER,XORR
		DW	QBRAN,BACK1
		DW	DOLIT,BKSPP,TECHO,ATEXE,DOLIT,1,SUBB
		DW	BLANK,TECHO,ATEXE
		DW	DOLIT,BKSPP,TECHO,ATEXE
BACK1:		DW	EXIT

;   TAP		( bot eot cur c -- bot eot cur )
;		Accept and echo the key stroke and bump the cursor.

		M_COLON	3,"TAP",TAP
		DW	DUPP,TECHO,ATEXE
		DW	OVER,CSTOR,DOLIT,1,PLUS,EXIT

;   kTAP	( bot eot cur c -- bot eot cur )
;		Process a key stroke, CR or backspace.
;		These key strokes are not in the range 32..127 decimal,
;		i.e. not printable ASCII.

		M_COLON	4,"kTAP",KTAP
		DW	DUPP,DOLIT,CRR,XORR		;check CR
		DW	QBRAN,KTAP2
		DW	DOLIT,BKSPP,XORR		;check backspace
		DW	QBRAN,KTAP1
		DW	BLANK,TAP,EXIT			;normal exit
KTAP1:		DW	BKSP,EXIT			;backspace exit
KTAP2:		DW	DROP,SWAP,DROP,DUPP,EXIT	;CR exit

;   accept	( b u -- b u )
;		Accept characters to input buffer. Return with actual count.

		M_COLON	6,"accept",ACCEP
		DW	OVER,PLUS,OVER
ACCP1:		DW	DDUP,XORR
		DW	QBRAN,ACCP4
		DW	KEY,DUPP
;kls;		DW	BLANK,SUBB,DOLIT,95,ULESS	;key(ascii)-32<95
		DW	BLANK,DOLIT,127,WITHI		;32<key(ascii)<127
		DW	QBRAN,ACCP2
		DW	TAP
		DW	BRAN,ACCP3
ACCP2:		DW	TTAP,ATEXE
ACCP3:		DW	BRAN,ACCP1
ACCP4:		DW	DROP,OVER,SUBB,EXIT

;   raccept	( b u -- b u )		( KLS_WORD )
;		Accept characters to input buffer. Return with actual count.

		M_COLON	7,"raccept",RACCEP
; OVER + OVER
		DW	OVER,PLUS,OVER	; inc buf ptr ( b u b+ u -- )
; BEGIN 2DUP XOR WHILE
RACCP1:		DW	DDUP,XORR	; ( b u b+ u b+ u -- )
		DW	QBRAN,RACCP4
		DW	ROMKEY		; get char from ROM instead!
; DUP BLANK 127 WITHIN ( .. f -- )
		DW	DUPP
;kls;		DW	BLANK,SUBB,DOLIT,95,ULESS	;key(ascii)-32<95
		DW	BLANK,DOLIT,127,WITHI		;32<key(ascii)<127
; IF TAP ELSE 'TAP @EXE THEN
		DW	QBRAN,RACCP2
		DW	TAP		; process printing chars
		DW	BRAN,RACCP3
RACCP2:		DW	TTAP,ATEXE	; kTAP process non-printing chars
; REPEAT
RACCP3:		DW	BRAN,RACCP1
RACCP4:		DW	DROP,OVER,SUBB	; adj buf ptr ( b u -- )
		DW	EXIT


;   EXPECT	( b u -- )
;		Accept input stream and store count in SPAN.

		M_COLON	6,"EXPECT",EXPEC
		DW	TEXPE,ATEXE,SPAN,STORE,DROP,EXIT

;   QUERY	( -- )
;		Accept input stream to terminal input buffer.

		M_COLON	5,"QUERY",QUERY
		DW	TIB,DOLIT,80,TEXPE,ATEXE,NTIB,STORE
		DW	DROP,DOLIT,0,INN,STORE,EXIT

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; Error handling									
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;   CATCH	( ca -- 0 | err# )
;		Execute word at ca and set up an error frame for it.

		M_COLON	5,"CATCH",CATCH
		DW	SPAT,TOR,HANDL,AT,TOR	;save error frame
		DW	RPAT,HANDL,STORE,EXECU	;execute
		DW	RFROM,HANDL,STORE	;restore error frame
		DW	RFROM,DROP,DOLIT,0,EXIT	;no error

;   THROW	( err# -- err# )
;		Reset system to current local error frame an update error flag.

		M_COLON	5,"THROW",THROW
		DW	HANDL,AT,RPSTO		;restore return stack
		DW	RFROM,HANDL,STORE	;restore handler frame
		DW	RFROM,SWAP,TOR,SPSTO	;restore data stack
		DW	DROP,RFROM,EXIT

;   NULL$	( -- a )
;		Return address of a null string with zero count.

		M_COLON	5,"NULL$",NULLS
		DW	DOVAR			;emulate CREATE
		DW	0
;kls;		DB	99,111,121,111,116,101	;c,o,y,o,t,e
		DW	0

;   ABORT	( -- )
;		Reset data stack and jump to QUIT.

		M_COLON	5,"ABORT",ABORT
		DW	NULLS,THROW

;   abort"	( f -- )
;		Run time routine of ABORT" . Abort with a message.

		M_COLON	COMPO+6,"abort\x22",ABORQ
		DW	QBRAN,ABOR1		;test flag
		DW	DOSTR,THROW		;pass error string
ABOR1:		DW	DOSTR,DROP,EXIT		;drop error

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; The text interpreter									
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;   $INTERPRET	( a -- )
;		Interpret a word. If failed, try to convert it to an integer.

		M_COLON	10,"$INTERPRET",INTER
		DW	NAMEQ,QDUP		;?defined
		DW	QBRAN,INTE1		;0 = no, try number @ INTE1
;kls;		DW	AT,DOLIT,COMPO,ANDD	;?compile only lexicon bits
		DW	CAT,DOLIT,COMPO,ANDD	;?compile only lexicon bits kls
		DW	ABORQ
		DB	15," compile only  "
		DW	EXECU,EXIT		;execute defined word
INTE1:		DW	TNUMB,ATEXE		;convert a number
		DW	QBRAN,INTE2
		DW	EXIT
INTE2:		DW	THROW			;error

;   [		( -- )
;		Start the text interpreter.

		M_COLON	IMEDD+1,"[",LBRAC
		DW	DOLIT,INTER,TEVAL,STORE,EXIT

;   .OK		( -- )
;		Display 'ok' only while interpreting.

		M_COLON	3,".OK",DOTOK
		DW	DOLIT,INTER,TEVAL,AT,EQUAL
		DW	QBRAN,DOTO1
		DW	DOTQP
		DB	3," ok"
DOTO1:		DW	CR,EXIT

;   ?STACK	( -- )
;		Abort if the data stack underflows.

		M_COLON	6,"?STACK",QSTAC
		DW	DEPTH,ZLESS		;check only for underflow
		DW	ABORQ
		DB	10," underflow",0
		DW	EXIT

;   EVAL	( -- )
;		Interpret the input stream.

		M_COLON	4,"EVAL",EVAL
EVAL1:		DW	TOKEN,DUPP,CAT		;?input stream empty
		DW	QBRAN,EVAL2
		DW	TEVAL,ATEXE,QSTAC	;evaluate input, check stack
		DW	BRAN,EVAL1
EVAL2:		DW	DROP,TPROM,ATEXE,EXIT	;prompt

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; Shell										
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;   PRESET	( -- )
;		Reset data stack pointer and the terminal input buffer.

		M_COLON	6,"PRESET",PRESE
		DW	SZERO,AT,SPSTO
		DW	DOLIT,TIBB,NTIB,CELLP,STORE,EXIT

;   xio		( a a a -- )
;		Reset the I/O vectors 'EXPECT, 'TAP, 'ECHO and 'PROMPT
;		in USER variables...
;	DFL	QRX			;'?KEY 			6
;	DFL	TXSTO			;'EMIT 			7
;	DFL	ACCEP			;'EXPECT 		8
;	DFL	KTAP			;'TAP	 		9
;	DFL	TXSTO			;'ECHO 			10
;	DFL	DOTOK			;'PROMPT 		11
		

		M_COLON	COMPO+3,"xio",XIO	; ( apr aec ata -- )
		DW	DOLIT,ACCEP		; ( apr aec ata aex -- )
		DW	TEXPE,DSTOR		; ( apr aec -- )
		DW	TECHO,DSTOR		; ( -- )
		DW	EXIT

;   FILE	( -- )
;		Select I/O vectors for file download.

		M_COLON	4,"FILE",FILE
		DW	DOLIT,PACE,DOLIT,DROP
		DW	DOLIT,KTAP,XIO,EXIT

;   HAND	( -- )
;		Select I/O vectors for terminal interface.

		M_COLON	4,"HAND",HAND
		DW	DOLIT,DOTOK,DOLIT,EMIT
		DW	DOLIT,KTAP,XIO,EXIT

;   ROMin	( -- )		(**KLS**)
;		Select I/O vectors for ROM input words

		M_COLON	5,"ROMin",ROMIN
		DW	DOLIT,PACE,DOLIT,DROP
		DW	DOLIT,KTAP
		DW	DOLIT,RACCEP		; simulate XIO  
		DW	TEXPE,DSTOR
		DW	TECHO,DSTOR	
		DW	EXIT

;   I/O		( -- a )
;		Array to store default I/O vectors.

		M_COLON	3,"I/O",ISLO
		DW	DOVAR			;emulate CREATE
		DW	QRX,TXSTO		;default I/O vectors

;   CONSOLE	( -- )
;		Initiate terminal interface.

		M_COLON	7,"CONSOLE",CONSO
		DW	ISLO,DAT,TQKEY,DSTOR	;restore default I/O device
		DW	HAND,EXIT		;keyboard input

;   QUIT	( -- )
;		Reset return stack pointer and start text interpreter.

		M_COLON	4,"QUIT",QUIT
		DW	RZERO,AT,RPSTO		;reset return stack pointer
QUIT1:		DW	LBRAC			;start interpretation
QUIT2:		DW	QUERY			;get input
;kls;		DW	TDUMP			;debug...TIB
		DW	DOLIT,EVAL,CATCH,QDUP	;evaluate input
		DW	QBRAN,QUIT2		;continue till error
		DW	TPROM,AT,TOR		;save input device
		DW	CONSO,NULLS,OVER,XORR	;?display error message
		DW	QBRAN,QUIT3
		DW	SPACE,COUNT,TYPES	;error message
		DW	DOTQP
		DB	3," ? "			;error prompt
QUIT3:		DW	RFROM,DOLIT,DOTOK,XORR	;?file input
		DW	QBRAN,QUIT4
		DW	DOLIT,ERR,EMIT		;file error, tell host
QUIT4:		DW	PRESE			;some cleanup
		DW	BRAN,QUIT1		;stay here

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; The compiler										
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;   '		( -- ca )
;		Search context vocabularies for the next word in input stream.

		M_COLON	1,"'",TICK
		DW	TOKEN,NAMEQ		;?defined
		DW	QBRAN,TICK1
		DW	EXIT			;yes, push code address
TICK1:		DW	THROW			;no, error

;   ALLOT	( n -- )
;		Allocate n bytes to the code dictionary.

		M_COLON	5,"ALLOT",ALLOT
		DW	CP,PSTOR,EXIT		;adjust code pointer

;   ,		( w -- )
;		Compile an integer into the code dictionary.


		M_COLON	1,"\x2c",COMMA
		DW	HERE,DUPP,CELLP		;cell boundary
		DW	CP,STORE,STORE,EXIT	;adjust code pointer and compile

;   [COMPILE]	( -- ; <string> )
;		Compile the next immediate word into code dictionary.

		M_COLON	IMEDD+9,"[COMPILE]",BCOMP
		DW	TICK,COMMA,EXIT

;   COMPILE	( -- )
;		Compile the next address in colon list to code dictionary.

		M_COLON	COMPO+7,"COMPILE",COMPI
		DW	RFROM,DUPP,AT,COMMA	;compile address
		DW	CELLP,TOR,EXIT		;adjust return address

;   LITERAL	( w -- )
;		Compile tos to code dictionary as an integer literal.

		M_COLON	IMEDD+7,"LITERAL",LITER
		DW	COMPI,DOLIT,COMMA,EXIT

;   $,"		( -- )
;		Compile a literal string up to next " .

		M_COLON	3,"$\x2c\x22",STRCQ		
		DW	DOLIT,'"',WORDD	;move string to code dictionary
		DW	COUNT,PLUS,ALGND	;calculate aligned end of string
		DW	CP,STORE,EXIT		;adjust the code pointer

;   RECURSE	( -- )
;		Make the current word available for compilation.

		M_COLON	IMEDD+7,"RECURSE",RECUR
		DW	LAST,AT,NAMET,COMMA,EXIT

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; Structures										
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;   FOR		( -- a )
;		Start a FOR-NEXT loop structure in a colon definition.

		M_COLON	IMEDD+3,"FOR",FOR
		DW	COMPI,TOR,HERE,EXIT

;   BEGIN	( -- a )
;		Start an infinite or indefinite loop structure.

		M_COLON	IMEDD+5,"BEGIN",BEGIN
		DW	HERE,EXIT

;   NEXT	( a -- )
;		Terminate a FOR-NEXT loop structure.

		M_COLON	IMEDD+4,"NEXT",NEXT
		DW	COMPI,DONXT,COMMA,EXIT

;   UNTIL	( a -- )
;		Terminate a BEGIN-UNTIL indefinite loop structure.

		M_COLON	IMEDD+5,"UNTIL",UNTIL
		DW	COMPI,QBRAN,COMMA,EXIT

;   AGAIN	( a -- )
;		Terminate a BEGIN-AGAIN infinite loop structure.

		M_COLON	IMEDD+5,"AGAIN",AGAIN
		DW	COMPI,BRAN,COMMA,EXIT

;   IF		( -- A )
;		Begin a conditional branch structure.

		M_COLON	IMEDD+2,"IF",IFF
		DW	COMPI,QBRAN,HERE
		DW	DOLIT,0,COMMA,EXIT

;   AHEAD	( -- A )
;		Compile a forward branch instruction.

		M_COLON	IMEDD+5,"AHEAD",AHEAD
		DW	COMPI,BRAN,HERE,DOLIT,0,COMMA,EXIT

;   REPEAT	( A a -- )
;		Terminate a BEGIN-WHILE-REPEAT indefinite loop.

		M_COLON	IMEDD+6,"REPEAT",REPEA
		DW	AGAIN,HERE,SWAP,STORE,EXIT

;   THEN	( A -- )
;		Terminate a conditional branch structure.

		M_COLON	IMEDD+4,"THEN",THENN
		DW	HERE,SWAP,STORE,EXIT

;   AFT		( a -- a A )
;		Jump to THEN in a FOR-AFT-THEN-NEXT loop the first time through.

		M_COLON	IMEDD+3,"AFT",AFT
		DW	DROP,AHEAD,BEGIN,SWAP,EXIT

;   ELSE	( A -- A )
;		Start the false clause in an IF-ELSE-THEN structure.

		M_COLON	IMEDD+4,"ELSE",ELSEE
		DW	AHEAD,SWAP,THENN,EXIT

;   WHILE	( a -- A a )
;		Conditional branch out of a BEGIN-WHILE-REPEAT loop.

		M_COLON	IMEDD+5,"WHILE",WHILE
		DW	IFF,SWAP,EXIT

;   ABORT"	( -- ; <string> )
;		Conditional abort with an error message.

		M_COLON	IMEDD+6,"ABORT\x22",ABRTQ
		DW	COMPI,ABORQ,STRCQ,EXIT

;   $"		( -- ; <string> )
;		Compile an inline string literal.

		M_COLON	IMEDD+2,"$\x22",STRQ
		DW	COMPI,STRQP,STRCQ,EXIT

;   ."		( -- ; <string> )
;		Compile an inline string literal to be typed out at run time.

		M_COLON	IMEDD+2,".\x22",DOTQ
		DW	COMPI,DOTQP,STRCQ,EXIT

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; Name compiler									
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;   ?UNIQUE	( a -- a )
;		Display a warning message if the word already exists.

		M_COLON	7,"?UNIQUE",UNIQU
		DW	DUPP,NAMEQ		;?name exists
		DW	QBRAN,UNIQ1
		DW	DOTQP			;redefinitions are OK
		DB	7," reDef "		;but the user should be warned
		DW	OVER,COUNT,TYPES	;just in case its not planned
UNIQ1:		DW	DROP,EXIT

;   $,n		( na -- )
;		Build a new dictionary name using the string at na.

		M_COLON	3,"$\x2cn",SNAME
		DW	DUPP,CAT		;?null input
		DW	QBRAN,PNAM1
		DW	UNIQU			;?redefinition. stack unchanged
		DW	DUPP,LAST,STORE		;save this na for vocabulary link
;	( na -- )

;; v1.x		DW	HERE,ALGND,SWAP		;align code address HERE = CP @
;; v1.x		DW	CELLM			;link address
;; v1.x		DW	CRRNT,AT,AT,OVER,STORE	;write LFA
;; v1.x		DW	CELLM,DUPP,NP,STORE	;adjust name pointer
;; v1.x		DW	STORE			;save code pointer

		DW	DUPP		; TING'S CODE DOES THIS. WHY?
		DW	DUPP			; ( na na na -- )
		DW	CAT			; get length byte of packed string
		DW	PLUS			; ( na na a -- ) a=na+length
		DW	DOLIT,1
		DW	PLUS			; ( na na ac -- ) ac=na+length+1
		DW	ALGND			; be sure address ac is aligned
		DW	DUPP			; ( na na ac ac -- )
		DW	CP			; ( na na ac ac cpa -- )
		DW	STORE			; save code pointer ( na na ac -- )
		DW	SWAP			; ( na ac na -- )
		DW	CELLM			; ( na ac al -- ) al=new word lfa
		DW	CRRNT,AT,AT		; last defined word in current vocab's LFA
		DW	OVER			; ( na ac al olfa al -- )
		DW	STORE			; write LFA ( na ac al -- )
		DW	CELLM,STORE		; write CFA ( na -- )

		DW	EXIT
PNAM1:		DW	STRQP
		DB	5," name",0,0		;null input
		DW	THROW

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; FORTH compiler									
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
		
;   $COMPILE	( a -- )
;		Compile next word to code dictionary as a token or literal.

		M_COLON	8,"$COMPILE",SCOMP
		DW	NAMEQ,QDUP		;?defined
		DW	QBRAN,SCOM2
;kls;		DW	AT,DOLIT,IMEDD,ANDD	;?immediate
		DW	CAT,DOLIT,IMEDD,ANDD	;?immediate kls
		DW	QBRAN,SCOM1
		DW	EXECU,EXIT		;its immediate, execute
SCOM1:		DW	COMMA,EXIT		;its not immediate, compile
SCOM2:		DW	TNUMB,ATEXE		;try to convert to number
		DW	QBRAN,SCOM3
		DW	LITER,EXIT		;compile number as integer
SCOM3:		DW	THROW			;error

;   OVERT	( -- )
;		Link a new word into the current vocabulary.

		M_COLON	5,"OVERT",OVERT
		DW	LAST,AT,CRRNT,AT,STORE,EXIT

;   ;		( -- )
;		Terminate a colon definition.

		M_COLON	IMEDD+COMPO+1,";",SEMIS
		DW	COMPI,EXIT,LBRAC,OVERT,EXIT

;   ]		( -- )
;		Start compiling the words in the input stream.

		M_COLON	1,"]",RBRAC
		DW	DOLIT,SCOMP,TEVAL,STORE,EXIT

;   call,	( ca -- )
;		Assemble a call instruction to ca.

		M_COLON	5,"call\x2c",CALLC
		DW	DOLIT,CALLL,COMMA	;Direct Threaded Code
		DW	COMMA,EXIT

;   :		( -- ; <string> )
;		Start a new colon definition using next word as its name.

		M_COLON	1,":",COLON
		DW	TOKEN,SNAME,DOLIT,DOLST
		DW	CALLC,RBRAC,EXIT

;   IMMEDIATE	( -- )
;		Make the last compiled word an immediate word.

		M_COLON	9,"IMMEDIATE",IMMED
;kls;		DW	DOLIT,IMEDD,LAST,AT,AT,ORR
		DW	DOLIT,IMEDD,LAST,AT,CAT,ORR	;kls
;kls;		DW	LAST,AT,STORE,EXIT
		DW	LAST,AT,CSTOR,EXIT		;kls

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; Defining words									
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;   USER	( u -- ; <string> )
;		Compile a new user variable.

		M_COLON	4,"USER",USER
		DW	TOKEN,SNAME,OVERT
		DW	DOLIT,DOLST,CALLC
		DW	DOLIT,DOUSE,COMMA
		DW	COMMA,EXIT

;   CREATE	( -- ; <string> )
;		Compile a new array entry without allocating code space.

		M_COLON	6,"CREATE",CREAT
		DW	TOKEN,SNAME,OVERT
		DW	DOLIT,DOLST,CALLC
		DW	DOLIT,DOVAR,COMMA,EXIT

;   VARIABLE	( -- ; <string> )
;		Compile a new variable initialized to 0.

		M_COLON	8,"VARIABLE",VARIA
		DW	CREAT,DOLIT,0,COMMA,EXIT

;   DOES>	( -- a )		(**KLS**)
;		Puts PFA of run time code on stack.
;		PFA of colon words points to list of addresses.

		M_COLON	IMEDD+5,"DOES>",DOEST
; save pfa of next word (after DOES>)
; find cfa of word definition DOES> appears in (latest dictionary entry)
;
; assembler example ...
; : 1MI ( opcode -- ) CREATE C, DOES> C@ C, ;
; 00 1MI NOP
;
		DW	COMPI,doDOES
		DW	COMPI,EXIT
		DW	DOLIT,DOLST,CALLC
		DW	COMPI,RFROM
		DW	EXIT

;  CONSTANT ( n -- ; -- n )	(**KLS**)
;		Compile a constant value

; : CONSTANT CREATE , DOES> @ ;

		M_COLON 8,"CONSTANT",CONST
		DW	CREAT,COMMA,doDOES,EXIT
; VER 1.x	M_COLON 10,"doCONSTANT",doCONST
; CAN'T HAVE NAME DICTIONARY ENTRY WHERE DOES CODE IS SUPPOSED TO BE
		DW	CALLL
		DW	DOLST
		DW	RFROM,AT,EXIT

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; Tools										
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;   _TYPE	( b u -- )
;		Display a string. Filter non-printing characters.

		M_COLON	5,"_TYPE",UTYPE
		DW	TOR			;start count down loop
		DW	BRAN,UTYP2		;skip first pass
UTYP1:		DW	DUPP,CAT,TCHAR,EMIT	;display only printable
		DW	DOLIT,1,PLUS		;increment address
UTYP2:		DW	DONXT,UTYP1		;loop till done
		DW	DROP,EXIT

;   dm+		( a u -- a )
;		Dump u bytes from , leaving a+u on the stack.
;		KLS - UDOTR -> UDOTR2 leading zero format fix

		M_COLON	3,"dm+",DUMPP
		DW	OVER,DOLIT,4,UDOTR	;display address
		DW	SPACE,TOR		;start count down loop
		DW	BRAN,PDUM2		;skip first pass
PDUM1:		DW	DUPP,CAT,DOLIT,3,UDOTR2	;display numeric data
		DW	DOLIT,1,PLUS		;increment address
PDUM2:		DW	DONXT,PDUM1		;loop till done
		DW	EXIT

;   DUMP	( a u -- )
;		Dump u bytes from a, in a formatted manner.

		M_COLON	4,"DUMP",DUMP
		DW	BASE,AT,TOR,HEX		;save radix, set hex
		DW	DOLIT,16,SLASH		;change count to lines
		DW	TOR			;start count down loop
DUMP1:		DW	CR,DOLIT,16,DDUP,DUMPP	;display numeric
		DW	ROT,ROT
		DW	DOLIT,2,SPACS,UTYPE	;display printable characters
		DW	NUFQ,INVER		;user control
		DW	QBRAN,DUMP2
		DW	DONXT,DUMP1		;loop till done
		DW	BRAN,DUMP3
DUMP2:		DW	RFROM,DROP		;cleanup loop stack, early exit
DUMP3:		DW	DROP,RFROM,BASE,STORE	;restore radix
		DW	EXIT

;   .S		( ... -- ... )
;		Display the contents of the data stack.

		M_COLON	2,".S",DOTS
		DW	CR,DEPTH		;stack depth
		DW	TOR			;start count down loop
		DW	BRAN,DOTS2		;skip first pass
DOTS1:		DW	RAT,PICK,DOT		;index stack, display contents
DOTS2:		DW	DONXT,DOTS1		;loop till done
		DW	DOTQP			;print tos string...
		DB	3,"<sp"
		DW	EXIT

;   !CSP	( -- )
;		Save stack pointer in CSP for error checking.

		M_COLON	4,"!CSP",STCSP
		DW	SPAT,CSP,STORE,EXIT	;save pointer

;   ?CSP	( -- )
;		Abort if stack pointer differs from that saved in CSP.

		M_COLON	4,"?CSP",QCSP
		DW	SPAT,CSP,AT,XORR	;compare pointers
		DW	ABORQ			;abort if different
		DB	6,"stacks",0
		DW	EXIT

;   >NAME	( ca -- na | F )
;		Convert code address to a name address.

		M_COLON	5,">NAME",TNAME
		DW	CRRNT			;vocabulary link
TNAM1:		DW	CELLP,AT,QDUP		;check all vocabularies
		DW	QBRAN,TNAM4
		DW	DDUP
TNAM2:		DW	AT,DUPP			;?last word in a vocabulary
		DW	QBRAN,TNAM3
		DW	DDUP,NAMET,XORR		;compare
		DW	QBRAN,TNAM3
		DW	CELLM			;continue with next word
		DW	BRAN,TNAM2
TNAM3:		DW	SWAP,DROP,QDUP
		DW	QBRAN,TNAM1
		DW	SWAP,DROP,SWAP,DROP,EXIT
TNAM4:		DW	DROP,DOLIT,0,EXIT

;   .ID		( na -- )
;		Display the name at address.

		M_COLON	3,".ID",DOTID
		DW	QDUP			;if zero no name
		DW	QBRAN,DOTI1
		DW	COUNT,DOLIT,01FH,ANDD	;mask lexicon bits
		DW	UTYPE,EXIT		;display name string
DOTI1:		DW	DOTQP
		DB	11," {noName}  "
		DW	EXIT

;   SEE		( -- ; <string> )
;		A simple decompiler.

		M_COLON	3,"SEE",SEE
		DW	TICK			;starting address
		DW	CR,CELLP
SEE1:		DW	CELLP,DUPP,AT,DUPP	;?does it contain a zero
		DW	QBRAN,SEE2
		DW	TNAME			;?is it a name
SEE2:		DW	QDUP			;name address or zero
		DW	QBRAN,SEE3
		DW	SPACE,DOTID		;display name
		DW	BRAN,SEE4
SEE3:		DW	DUPP,AT,UDOT		;display number
SEE4:		DW	NUFQ			;user control
		DW	QBRAN,SEE1
		DW	DROP,EXIT

;   SEES	( -- ; <string> )
;		A simple decompiler that knows when to stop.
;		Stops when the next instruction is CALLN -> ( NOP JSR )

		M_COLON	4,"SEES",SEES
		DW 	TICK
		DW	CR
		DW	CELLP
		DW	CELLP
		DW	DUPP
		DW	AT
		DW	DUPP
		DW 	QBRAN,SEES+0018H	
		DW	TNAME
		DW	QDUP
		DW	QBRAN,SEES+00026H
		DW	SPACE
		DW	DOTID
		DW	BRAN,SEES+0004CH
		DW	DUPP
		DW	AT
		DW	doLIT
		DW	CALLL
		DW 	EQUAL
		DW	OVER
		DW	HERE
		DW	CELLP
		DW	EQUAL
		DW	ORR
		DW	QBRAN,SEES+000046H
		DW	DROP
		DW	EXIT
		DW	BRAN,SEES+00004CH
		DW	DUPP
		DW	AT
		DW	UDOT
		DW	NUFQ
		DW	QBRAN,SEES+0000AH
		DW	DROP
		DW	EXIT
                               


;   WORDS	( -- )
;		Display the names in the context vocabulary.

		M_COLON	5,"WORDS",WORDS
		DW	CR,CNTXT,AT		;only in context
WORS1:		DW	AT,QDUP			;?at end of list
		DW	QBRAN,WORS2
		DW	DUPP,SPACE,DOTID	;display a name
		DW	CELLM,NUFQ		;user control
		DW	QBRAN,WORS1
		DW	DROP
WORS2:		DW	EXIT


;   TDUMP	( -- )
;		Dump TIB

	M_COLON	5,"TDUMP",TDUMP
	DW	TIB
	DW	DOLIT
	DW	80
	DW	DUMP		;dump TIB
	DW	EXIT



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; Hardware reset									
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;   VER		( -- n )
;		Return the version number of this implementation.

		M_COLON	3,"VER",VERSN
		DW	DOLIT,VER*256+EXT,EXIT

;   hi		( -- )
;		Display the sign-on message of eForth.

		M_COLON	2,"hi",HI
		DW	STOIO,CR,DOTQP		;initialize I/O
		DB	13,"eForth v",(VER+'0'),'.',(EXT+'0'),ALFA,0	;model+version+alpha
		DW	CR,EXIT

;   'BOOT	( -- a )
;		The application startup vector.

		M_COLON	5,"'BOOT",TBOOT
		DW	DOVAR
		DW	HI			;application to boot

	IF DIAGS
	INCL "E4DIAG.ASM"
	ENDIF

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; KERNEL EXTENSIONS HERE								
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; VOCABULARY defines a new vocabulary by (compile time:) creating a link
; to the most recently defined word in the vocabulary (initially null)
; and a link to the previously defined vocabulary. At run time:
; the address of the link structure is stored in CONTEXT.

;; : VOCABULARY 
;;	CREATE HERE ( put PFA on stack )
;;	name-link , ( initially 0 )
;;	voc-link @ , ( save old voc-link )
;;	voc-link !	( new voc-link ) 
;;	DOES> CONTEXT ! ;


     M_COLON  10,"VOCABULARY",VOCABULARY
    DW CREAT
    DW HERE
    DW DOLIT,0
    DW COMMA
    DW VOCLINK
    DW AT
    DW COMMA
    DW VOCLINK
    DW STORE
    DW doDOES
    DW EXIT

; v1.x 
;    M_COLON 12,"doVOCABULARY",doVOCABULARY
; CAN'T HAVE NAME DICTIONARY ENTRY WHERE DOES CODE IS SUPPOSED TO BE

doVOC:		; but an assembler label is OK

	DW	CALLL
	DW	DOLST

    DW RFROM
    DW CNTXT
    DW STORE
    DW EXIT



; ALSO pushes the vocabulary stack and duplicates the top vocabulary.
; hard code the stack size - 1 (=7).

;; : ALSO 
;; 6 FOR CONTEXT ( n..0 n+1 iterations )
;; R@ 1 + ( never do 0 = CONTEXT address )
;; 4 * + ( calculate vocabulary stack address )
;; DUP CELL- ( vocabulary stack address -1 )
;; @ SWAP ! 
;; NEXT ;

     M_COLON  4,"ALSO",ALSO
    DW DOLIT,6
    DW TOR
    DW CNTXT
    DW RAT
    DW DOLIT,1,PLUS
    DW DOLIT,4
    DW STAR
    DW PLUS
    DW DUPP
    DW CELLM
    DW AT
    DW SWAP
    DW STORE
    DW DONXT,ALSO+000AH
    DW EXIT


; ONLY flushes the vocabulary stack leaving FORTH on the top.

;; : ONLY 
;; FORTH ( new context FORTH )
;; 6 FOR CONTEXT ( n..0 n+1 iterations )
;; R@ 4 * + ( calculate vocabulary stack address )
;; CELL+ ( offset one cell so 0 index doesn't clear FORTH )
;; 0 SWAP ! ( clear stack entry )
;; NEXT ;

     M_COLON  4,"ONLY",ONLY
    DW FORTH
    DW DOLIT,6
    DW TOR
    DW CNTXT
    DW RAT
    DW DOLIT,4
    DW STAR
    DW PLUS
    DW CELLP
    DW DOLIT,0
    DW SWAP
    DW STORE
    DW DONXT,ONLY+000CH
    DW EXIT


; DEFINITIONS copies the content of CONTEXT to CURRENT

;; : DEFINITIONS CONTEXT @ CURRENT ! ;

     M_COLON  11,"DEFINITIONS",DEFINITIONS
    DW CNTXT
    DW AT
    DW CRRNT
    DW STORE
    DW EXIT

; ORDER displays vocabulary order in CONTEXT 
; and displays vocabulary in CURRENT
;
; : ORDER CR 7 FOR CONTEXT R@ 
; 7 SWAP - 4 * + @
; CELL- CELL- CELL- >NAME .ID CR
; NEXT ;

    M_COLON  5,"ORDER",ORDER
    DW CR
    DW DOLIT,7
    DW TOR
    DW CNTXT
    DW RAT
    DW DOLIT,7
    DW SWAP
    DW SUBB
    DW DOLIT,4
    DW STAR
    DW PLUS
    DW AT
    DW CELLM
    DW CELLM
    DW CELLM
    DW TNAME
    DW DOTID
    DW CR
    DW DONXT,ORDER+000CH
    DW EXIT



; VOCS lists the vocabularies in the vocabulary linked list

;; : VOCS voc-link BEGIN CR ( voc-link var addr = pointer )
;; @ DUP CELL+ SWAP ( link duped; next pointer )
;; CELL- CELL- CELL- ( name addr of this voc )
;; >NAME .ID DUP @ ( display name; dup link )
;; 0 = UNTIL ( forth is always the last in the list )
;; CR DROP ;

     M_COLON  4,"VOCS",VOCS
    DW VOCLINK
    DW CR
    DW AT
    DW DUPP
    DW CELLP
    DW SWAP
    DW CELLM
    DW CELLM
    DW CELLM
    DW TNAME
    DW DOTID
    DW DUPP
    DW AT
    DW DOLIT,0
    DW EQUAL
    DW QBRAN,VOCS+0006H
    DW CR
    DW DROP
    DW EXIT



;; BLOAD ( a --  )
;; Binary Load file to buffer at address a
    M_COLON  5,"BLOAD",BLOAD
    DW DUPP
    DW DOLIT
    DW 00001H
    DW PLUS
    DW SWAP
    DW QRX
    DW QBRAN,BLOAD+0000EH
    DW SWAP
    DW CSTOR
    DW DUPP
    DW DOLIT
    DW 00001H
    DW PLUS
    DW SWAP
    DW DOLIT
    DW 0FFF0H
    DW TOR
    DW QRX
    DW QBRAN,BLOAD+00040H
    DW DOLIT
    DW 0FFFFH
    DW RFROM
    DW DROP
    DW DOLIT
    DW 00000H
    DW TOR
    DW BRAN,BLOAD+00050H
    DW RAT
    DW DOLIT
    DW 00000H
    DW EQUAL
    DW QBRAN,BLOAD+00050H
    DW DOLIT
    DW 00000H
    DW DONXT,BLOAD+00028H
    DW QBRAN,BLOAD+00064H
    DW SWAP
    DW CSTOR
    DW DOLIT
    DW 00000H
    DW BRAN,BLOAD+0006CH
    DW DROP
    DW DROP
    DW DOLIT
    DW 0FFFFH
    DW QBRAN,BLOAD+00018H
    DW EXIT


;; READ ( a -- )
;; READS (compiles) the contents of the buffer at address a

	M_COLON	4,"READ",READ
	DW	RKINIT
	DW	ROMIN
	DW	EXIT

;; LOAD ( -- )
;; LOAD file using BLOAD then READ

	M_COLON	4,"LOAD",LOAD
	DW	HERE
	DW	DOLIT
	DW	0200H
	DW	PLUS
	DW	DUPP	
	DW	BLOAD
	DW	READ
	DW	EXIT
	

    M_COLON  6,"DOINIT",DOINIT
; DOINIT TOKEN NAME? IF EXECUTE EXIT THEN DROP ;
; if word name exists in dictionary, then execute it
; ( name -- )
    DW TOKEN
    DW NAMEQ
    DW QBRAN,DOINIT+0010H
    DW EXECU
    DW EXIT
    DW DROP
    DW EXIT



;; KLS DATA TABLE FOR ROMin
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;;	ROM DEFINITIONS THAT GET PROCESSED DURING COLD BOOT			
;;	Essentially an *.ini file.						
;;	WARNING: the compiler is slow.						
;;	MUST end by resetting vectors to the HANDshake mode!!!		
;;									
;;	INVOKE:								
;;	  ' ROM_DEF_NAME RKINIT	 ( e.g. ' ROMASSEM RKINIT )		
;;	  ROMin								
;;									
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	M_CODE	7,"ROMDEFS",ROMDEFS	
; add .ini definitions here...
;; insert source here...
;;	DB ": ( forth ) ;",CRR,LF
;; use ",22H," instead of "
	DB	"DOINIT INIT",CRR,LF
	DB	"HAND",CRR,LF		; THIS LINE MUST EXIST AND BE LAST


; RKINIT ( a -- )		KLS_WORD
;		Clears index used in reading from ROM

		M_CODE	6,"RKINIT",RKINIT
		POP 	HL		;pop start address
		DB 0DDH,07DH	; LD	IX,HL		; IX index = start addr
					; new Rabbit 2000 instruction (kls)
		M_NEXT

; ROMKEY: ( -- c )	KLS_WORD
;		Gets a character from ROM
;		Called by raccept

		M_CODE	6,"ROMKEY",ROMKEY
		LD	A,(IX+0)			; get byte
		INC	IX			; increment index
		LD	L,A
		LD	H,0
		PUSH	HL			; push c
		M_NEXT


		M_ALIGNC		; RE-ALIGN AFTER CODE


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;											
;; END KERNEL EXTENSIONS.								
;;											
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
                                        

;   COLD	( -- )
;		The hilevel cold start sequence.

		M_COLON	4,"COLD",COLD
COLD1:		
;;		DW	bpt			; breakpoint for testing
		DW	DOLIT,UZERO,DOLIT,UPP
		DW	DOLIT,ULAST-UZERO,CMOVE	;initialize user area
		DW	PRESE			;initialize data stack and TIB
	IF DIAGS
	DW	DIAG	;do diagnostics
	ENDIF
		DW	TBOOT,ATEXE		;application boot
		DW	FORTH			; sets CONTEXT = FORTH
		DW	CNTXT,AT,DUPP		
		DW	CRRNT,DSTOR,OVERT	;initialize search order
		DW	DOLIT,ROMDEFS,RKINIT	; KLS init index & pointer for ROMin
		DW	ROMIN			; KLS set vectors to get input from ROM
		DW	QUIT			;start interpretation
		DW 	BRAN,COLD1		;just in case

;===============================================================

LASTN		EQU	LINK		;last name address in ROM name dictionary
					;used to link new words to existing (ROM)

NTOP		EQU	TIBB-01000H	;next available memory in RAM name dictionary
					;NTOP NOT USED ver 2.x
					;(still has space in user var table)

CTOP		EQU	$		;next available memory in RAM code dictionary


;===============================================================


	END


